# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

# ****************************************  EpisodicMemoryMechanism ****************************************************

"""

Contents
--------

  * `EpisodicMemoryMechanism_Overview`
  * `EpisodicMemoryMechanism_Creation`
      - `EpisodicMemoryMechanism_Creation_InputPorts`
      - `EpisodicMemoryMechanism_Creation_Function_Parameters`
      - `EpisodicMemoryMechanism_Creation_OutputPorts`
  * `EpisodicMemoryMechanism_Structure`
      - `EpisodicMemoryMechanism_Memory_Fields`
      - `EpisodicMemoryMechanism_Input`
      - `EpisodicMemoryMechanism_Function`
      - `EpisodicMemoryMechanism_Output`
  * `EpisodicMemoryMechanism_Execution`
  * `EpisodicMemoryMechanism_Class_Reference`

.. _EpisodicMemoryMechanism_Overview:

Overview
--------

An EpisodicMemoryMechanism is a `ProcessingMechanism` that can store and retrieve items from a content addressable
memory;  that is, on each execution it can store an item presented to it as input, and use that to retrieve an item
from its memory based on the content of the input.  The `MemoryFunction` assigned as its `function
<EpisodicMemoryMechanism.function>` determines how items are stored and retrieved. Each memory is a list or array
composed of items referred to as `memory fields <EpisodicMemoryMechanism_Memory_Fields>`, each of which is a list or
array. Memories can have an arbitrary number of fields, and each of those can be of arbitrary shape, however all
memories for a given instance of an  EpisodicMemoryMechanism must have the same shape (number of fields, and shapes
of corresponding fields).  Each `InputPort` of an EpisodicMemoryMechanism provides the input for a corresponding
field of a memory to be stored and used for retrieval.  By default, each `OutputPort` contains the value of a field
of the last retrieved memory although, as with any Mechanism, OutputPorts can be `configured in other ways
<OutputPort_Customization>`. The full set of stored memories can be accessed from the Mechanism's `memory
<EpisodicMemoryMechanism.memory>` attribute, which references its `function's <EpisodicMemoryMechanism.function>`
memory `Parameter`. Other Parameters of its function (e.g., that regulate the probability of storage and/or retrieval
-- see `ContentAddressableMemory`) can be accessed and/or `modulated <ModulatorySignal_Modulation>` in the standard way
for a Mechanism's `function <EpisodicMemoryMechanism.function>`.

At present, EpisodicMemoryMechanism supports the following two MemoryFunctions:

• `ContentAddressableMemory` -- a general form of memory, that stores and retrieves memories as described above;
  any (weighted) combination of its fields can be used for retrieval.

• `DictionaryMemory` -- a more specific form of memory, that has only two fields, for keys and values, that stores
  these as pairs;  retrieval is based on similarity to the key;  this implements a format commonly used by applications
  that use dictionaries as a form of external memory.


.. _EpisodicMemoryMechanism_Creation:

Creating an EpisodicMemoryMechanism
-----------------------------------

.. _EpisodicMemoryMechanism_Creation_InputPorts:

*InputPorts, Entries and Memory Fields*
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

An EpisodicMemoryMechanism is created by calling its constructor with a specification of its `MemoryFunction` and
corresponding information about the number and shape of its `fields <EpisodicMemoryMechanism_Memory_Fields>`. These
are specified using the **default_variable** or **size** arguments of the constructor, in the standard way that the
`variable <Mechanism_Base.variable>` is specified for a `Component` (see `default_variable <Component_Variable>`, `size
<Component_Size>`).  These specify the shape of an entry `memory <EpisodicMemoryMechanism.memory>`, and used to create
the `InputPorts <InputPort>` for the Mechanism (see `Mechanism Variable <Mechanism_Variable_and_InputPorts>` for
additional information), with the number created equal to the number of items in the array (corresponding to fields in
an entry).  Each `input_port <EpisodicMemoryMechanism.input_ports>` provides the value assigned to a corresponding field
of the entry stored in `memory <EpisodicMemoryMechanism.memory>`, and used to retrieve one similar to it. By default,
`input_port <EpisodicMemoryMechanism.input_ports>` are named *FIELD_n_INPUT*, where "n" is replaced by the index of
each field; however, they can be named explicitly by specifying a list of strings in the **input_ports** argument of
the constructor; the number of these must equal the number of fields specified in **default_variable** or **size**.

.. _EpisodicMemoryMechanism_Creation_Function_Parameters:

*Function Parameters
~~~~~~~~~~~~~~~~~~~~

Parameters that govern the storage and retrieval process are specified as arguments to the `MemoryFunction` specified
in the **function** of the constructor (for example, see `ContentAddressableMemory` for parameters of the default
`function <EpisodicMemoryMechanism.function>`).

.. _EpisodicMemoryMechanism_Creation_OutputPorts:

*OutputPorts*
~~~~~~~~~~~~~

By default, a number of `OutputPorts <OutputPort>` is created equal to the number of InputPorts, each named either
*RETRIEVED_FIELD_n* or *RETRIEVED_<user specified InputPort name>*, that receive the values of the corresponding
fields of a retrieved memory.  OutputPort names can be specified explicitly, by assigning a list of strings to the
**output_ports** argument of the Mechanism's constructor;  however, in that case, or if any other forms of `OutputPort
specification <OutputPort_Forms_of_Specification>` are assigned to **output_ports**, then only the number of OutputPorts
specified are created, which may not match the number of fields of a retrieved memory.

.. _EpisodicMemoryMechanism_Structure:

Structure
---------

.. _EpisodicMemoryMechanism_Memory_Fields:

*Memory Fields*
~~~~~~~~~~~~~~~

Entries in the `memory <EpisodicMemoryMechanism.memory>` of an EpisodicMemoryMechanism are comprised of fields: lists
or 1d arrays within the outer list or array that comprise each entry. An entry can have an arbitrary number of fields,
and fields can be of arbitrary length.  However, all entries must have the same form (i.e., number of fields and shape
of corresponding fields across entries). One InputPort of the EpisodicMemoryMechanism is assigned to each field. Thus,
fields can be used to store different types of information in each field, and to retrieve entries from memory based on
all fields, or a weighted combination of them (as determined by the `MemoryFunction` assigned to `function
<EpisodicMemoryMechanism.function>`).

.. _EpisodicMemoryMechanism_Shape:

.. technical_note::

   The shape of an entry in `memory <EpisodicMemoryMechanism.memory>` is determined by the shape of the Mechanism's
   `variable <Mechanism_Base.variable>`. specified in the **default_variable** or **size** arguments of its constructor
   (see `EpisodicMemoryMechanism_Creation`).  Each item of `variable <Mechanism_Base.variable>` corresponds to a field.
   Both `memory <EpisodicMemoryMechanism.memory>` and all entries are stored in the EpisodicMemoryMechanism's `function
   <EpisodicMemoryMechanism.function>` as np.ndarrays, the dimensionality of which is determined by the shape of an
   entry and its fields. Fields are always stored as 1d arrays; if all fields have the same length (regular), then
   entries are 2d arrays and `memory <EpisodicMemoryMechanism.memory>` is a 3d array.  However, if fields have
   different lengths (`ragged <https://en.wikipedia.org/wiki/Jagged_array>`_) then, although each field is 1d, an
   entry is also 1d (with dtype='object'), and `memory <EpisodicMemoryMechanism.memory>` is 2d (with dtype='object').

.. _EpisodicMemoryMechanism_Input:

*Input*
~~~~~~~

An EpisodicMemoryMechanism has one or more `input_ports <EpisodicMemoryMechanism.input_ports>` that receive the
entry to be stored and that is used to retrieve an existing entry from its memory.  If the Mechanism is assigned
`ContentAddressableMemory` as its `function <EpisodicMemoryMechanism.function>`, then it can have an arbitrary
number of InputPorts, the input to which is assigned to the corresponding `memory field
<EpisodicMemoryMechanism_Memory_Fields>` of that function. By default InputPorts are named *FIELD_n_INPUT* (see
`EpisodicMemoryMechanism_Creation`). If the Mechanism is assigned `DictionaryMemory` as its `function
<EpisodicMemoryMechanism.function>`, then it is assigned at least one InputPort (named *KEY_INPUT* by default),
and optionally a second (named *VALUE_INPUT*) if **default_variable** or **size** specifies two items; any additional
fields are ignored.

.. _EpisodicMemoryMechanism_Function:

*Function*
~~~~~~~~~~

The default function is `ContentAddressableMemory` that can store entries with an arbitrary number of fields and
shapes, and retrieve them based on a weighted similarity to any combination of those fields.  A `DictionaryMemory`
can also be be assigned, that implements a more specific form of memory in which entries are made of up key-value
pairs, and retrieved based on similarity only to the key.  A custom function can also be specified, so long as it
meets the following requirements:

    * it must accept a list or array as its first argument, the items of which are lists or arrays;

    * it must return a list or array of the same size and shape as the input;

    * it must implement a ``memory`` attribute, that can be accessed by the EpisodicMemoryMechanism's `memory
      <EpisodicMemoryMechanism.memory>` attribute.

.. _EpisodicMemoryMechanism_Output:

*Output*
~~~~~~~~

By default, an EpisodicMemoryMechanism has a number of `OutputPorts <OutputPort>` equal to its number of InputPorts,
each of which is assigned the value of a corresponding field of the entry retrieved from memory (see
`EpisodicMemoryMechanism_Creation` for naming). However, if OutputPorts were specified in the constructor, then there
may be a different number of OutputPorts than InputPorts and `memory fields <EpisodicMemoryMechanism_Memory_Fields>`,
and thus some of the latter may not be reflected in any of the Mechanism's `output_ports
<EpisodicMemoryMechanism.output_ports>`. If the `function <EpisodicMemoryMechanism.function>` is a `DictionaryMemory`,
then it will have at least one OutputPort, named *KEY_OUTPUT*, that is assigned the key (first field) of the entry
retrieved from memory and, if two fields are specified in **default_variable** or **sze**, the Mechanism will have a
second OutputPort named *VALUE_OUTPUT* that is assigned the value (second field) of the entry retrieved from memory;
any additional ones are ignored and no other OutputPorts are created.

.. _EpisodicMemoryMechanism_Execution:

Execution
---------

When an EpisodicMemoryMechanism is executed, its `function <EpisodicMemoryMechanism.function>` carries out
the following operations:

    * retrieve an item from `memory <EpisodicMemoryMechanism.memory>` based on the `value <InputPort.value>` of
      its `input_ports <EpisodicMemoryMechanism.input_ports>`; if no retrieval is made, then an appropriately
      shaped zero-valued array is returned.
    ..
    * store the `value <InputPort.value>` of its `input_ports <EpisodicMemoryMechanism.input_ports>` as an entry in
      <EpisodicMemoryMechanism.function>` memory.
    ..
    * assign the value of the entry retrieved to its `output_ports <EpisodicMemoryMechanism.output_ports>`,
      based on how the latter are configured (see `EpisodicMemoryMechanism_Creation_OutputPorts`).

    .. note::
         In general, retrieval is executed before storage, so that the current items is not also retrieved;
         however, the order of storage and retrieval is determined by the EpisodicMemoryMechanism's
         `function <EpisodicMemoryMechanism.function>`.

.. _EpisodicMemoryMechanism_Class_Reference:

Class Reference
---------------


"""
from typing import Optional, Union

import numpy as np

from psyneulink.core.components.functions.function import Function
from psyneulink.core.components.functions.statefulfunctions.memoryfunctions import \
    DictionaryMemory, ContentAddressableMemory
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism_Base
from psyneulink.core.components.ports.inputport import InputPort
from psyneulink.core.globals.keywords import NAME, OWNER_VALUE, VARIABLE
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.globals.preferences.basepreferenceset import is_pref_set
from psyneulink.core.globals.utilities import deprecation_warning, convert_to_np_array

__all__ = ['EpisodicMemoryMechanism', 'KEY_INPUT', 'VALUE_INPUT', 'KEY_OUTPUT', 'VALUE_OUTPUT']

KEY_INPUT = 'KEY_INPUT'
VALUE_INPUT = 'VALUE_INPUT'
KEY_OUTPUT = 'KEY_OUTPUT'
VALUE_OUTPUT = 'VALUE_OUTPUT'
DEFAULT_INPUT_PORT_NAME_PREFIX = 'FIELD_'
DEFAULT_INPUT_PORT_NAME_SUFFIX = '_INPUT'
DEFAULT_OUTPUT_PORT_PREFIX = 'RETREIVED_'


class EpisodicMemoryMechanismError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


class EpisodicMemoryMechanism(ProcessingMechanism_Base):
    """
    EpisodicMemoryMechanism()

    Subclass of `ProcessingMechanism <ProcessingMechanism>` that implements a content addressable dictionary.
    See `Mechanism <Mechanism_Class_Reference>` for arguments and attributes.

    Attributes
    ----------

    input_ports : ContentAddressableList[str, InputPort]
        a list of the Mechanism's `InputPorts <Mechanism_InputPorts>`, the number of which is equal to the
        number of `memory fields <EpisodicMemoryMechanism_Memory_Fields>` in an entry of the Mechanism's `memory
        <EpisodicMemoryMechanism.memory>`, and that are named *FIELD_n_INPUT* by default (see
        `EpisodicMemoryMechanism_Input` and Mechanism `input_ports <Mechanism_Base.input_ports>` for additional
        information).

    function : MemoryFunction
        takes the Mechanism's `variable <Mechanism_Base.variable>` and uses it as a cue to retrieve an entry from
        `memory <EpisodicMemoryMechanism_Function>` based on its distance from existing entries, and then stores
        `variable <Mechanism_Base.variable>` in `memory <EpisodicMemoryMechanism_Function>` (see
        `EpisodicMemoryMechanism_Function` and Mechanism `function <Mechanism_Base.function>` for
        additional information).

    memory : 3d array
        contains entries stored in the `function <EpisodicMemoryMechanism.function>'\\s `memory` attribute.

    output_ports : ContentAddressableList[str, OutputPort]
        a list of the Mechanism's `OutputPorts <Mechanism_OutputPorts>`, the number of which is, by default, equal to
        the number of `memory fields <EpisodicMemoryMechanism_Memory_Fields>` in an entry of the Mechanism's `memory
        <EpisodicMemoryMechanism.memory>`, and each of which is assigned the value of the corresponding field of the
        last entry retrieved from `memory <EpisodicMemoryMechanism.memory>`.  However, as with any Mechanism, the
        number and value of OutputPorts can be customized (see `EpisodicMemoryMechanism_Output` and Mechanism
        `output_ports <Mechanism_Base.output_ports>` for additional information).
    """

    class Parameters(ProcessingMechanism_Base.Parameters):
        """
            Attributes
            ----------

                variable
                    see `variable <EpisodicMemoryMechanism.variable>`

                    :default value: [[0]]
                    :type: ``list``

                function
                    see `function <EpisodicMemoryMechanism.function>`

                    :default value: `DictionaryMemory`
                    :type: `Function`

        """
        variable = Parameter([[0]], pnl_internal=True, constructor_argument='default_variable')
        function = Parameter(ContentAddressableMemory, stateful=False, loggable=False)

    def __init__(self,
                 default_variable:Union[int, list, np.ndarray]=None,
                 size:Optional[Union[int, list, np.ndarray]]=None,
                 function:Optional[Function]=None,
                 params=None,
                 name=None,
                 prefs:is_pref_set=None,
                 **kwargs):

        # LEGACY SUPPORT FOR DictionaryMemory
        self._dictionary_memory = (isinstance(function, DictionaryMemory)
                                  or (isinstance(function, type)
                                      and function.__name__ is DictionaryMemory.__name__))
        if self._dictionary_memory:
            # Identify and warn about any deprecated args, and return their values for reassignment
            deprecated_arg_values = deprecation_warning(self, kwargs, {'content_size':'size'})
            # Assign value of deprecated args to current ones
            if 'size' in deprecated_arg_values:
                size = deprecated_arg_values['size']
            # Need to handle assoc_size specially, since it needs to be added to what was content_size
            if 'assoc_size' in kwargs:
                if isinstance(size, int):
                    size = [size,kwargs['assoc_size']]
                else:
                    size += kwargs['assoc_size']
                kwargs.pop('assoc_size')

        super().__init__(
            default_variable=default_variable,
            size=size,
            function=function,
            params=params,
            name=name,
            prefs=prefs,
            # input_ports=input_ports,
            # output_ports=output_ports,
            **kwargs
        )

    def _instantiate_input_ports(self, context=None):

        # FIX: REMOVE THIS AND DELETE "ELSE" WHEN DictionaryMemory IS RETIRED
        if self._dictionary_memory:
            input_ports = [KEY_INPUT]
            if len(self.parameters.variable.default_value) == 2:
                input_ports.append(VALUE_INPUT)
        else:
            input_ports = self.input_ports or [f'{DEFAULT_INPUT_PORT_NAME_PREFIX}{i}{DEFAULT_INPUT_PORT_NAME_SUFFIX}'
                                               for i in range(len(self.parameters.variable.default_value))]

        super()._instantiate_input_ports(input_ports=input_ports, context=context)

    def _instantiate_output_ports(self, context=None):
        """Generate OutputPorts with names specified and values with shapes equal to corresponding InputPorts

        If OutputPorts have not been specified, use InputPort names with prefix replaced and suffix removed.

        If any OutputPorts are specified as strings, those are used as names and are the only OutputPorts instantiated
            (even if the number is less than the number of InputPorts)

        If any OutputPorts are specified in a form other than a string, then this method is ignored and OutputPorts
            are instantiated in call to super().instantiate_output_ports;
            note:  in that case, the shapes of the value are as specified and may not necessarily correspond to the
                   shapes of the corresponding Inputs (i.e., memory fields).
        """

        # FIX: REMOVE THIS AND DELETE "ELSE" WHEN DictionaryMemory IS RETIRED
        if self._dictionary_memory:
            output_ports = [{NAME: KEY_OUTPUT,
                            VARIABLE: (OWNER_VALUE, 0)}]
            if len(self.parameters.variable.default_value) == 2:
                output_ports.append({NAME: VALUE_OUTPUT,
                                     VARIABLE: (OWNER_VALUE, 1)})
            self.parameters.output_ports._set(output_ports, override=True, context=context)

        else:
            output_ports_spec = self.output_ports
            # If output_ports was not specified or they are all strings, instantiate
            # (note: need to instantiate here if output_port specs are strings, to be sure values are correct,
            #    as this doesn't seem to be handled properly by super()._instantiate_output_ports)
            if output_ports_spec is None or all(isinstance(o, str) for o in output_ports_spec):
                output_ports = []
                input_suffix_len = len(DEFAULT_INPUT_PORT_NAME_SUFFIX)
                # Total number should be either the number of names provided, or default to length of Mech's value:
                num_output_ports = len(output_ports_spec) if output_ports_spec else len(self.value)
                for i in range(num_output_ports):
                    # Names specified, so use those:
                    if output_ports_spec:
                        output_ports.append({NAME: self.output_ports[i],
                                             VARIABLE: (OWNER_VALUE, i)})
                    # Otherwise, use InputPort names as base, removing DEFAULT_INPUT_PORT_NAME_SUFFIX
                    else:
                        input_port_name = self.input_ports[i].name
                        # if input_port_name[-input_suffix_len:] == DEFAULT_INPUT_PORT_NAME_SUFFIX:
                        # if not self.input_ports[i]._user_specified:
                        if not self.parameters.input_ports._user_specified:
                            input_port_name = input_port_name[:-input_suffix_len]
                        output_ports.append({NAME: DEFAULT_OUTPUT_PORT_PREFIX + input_port_name,
                                             VARIABLE: (OWNER_VALUE, i)})
                self.parameters.output_ports._set(output_ports, override=True, context=context)

        super()._instantiate_output_ports(context=context)

    # IMPLEMENTATION NOTE: REMOVE THIS METHOD WHEN DictionaryMemory IS RETIRED
    def _parse_function_variable(self, variable, context=None):

        if self._dictionary_memory:
            # If assoc has not been specified, add empty list to call to function (which expects two items in its variable)
            if len(variable) != 2:
                return convert_to_np_array([variable[0],[]])
            else:
                # Check that both are assigned inputs:
                missing_inputs = [self.input_ports.names[i] for i,t in enumerate([v for v in variable]) if t is None]
                if missing_inputs:
                    if len(missing_inputs) == 1:
                        missing_str = 'an input'
                        s = ''
                    else:
                        missing_str = 'inputs'
                        s = 's'
                    raise EpisodicMemoryMechanismError(f"{self.name} is missing {missing_str} for its"
                                                       f" {'and '.join(missing_inputs)} {InputPort.__name__}{s}.")
        return variable

    # IMPLEMENTATION NOTE: REMOVE THIS METHOD WHEN DictionaryMemory IS RETIRED
    def _execute(self, variable=None, context=None, runtime_params=None):

        value = super()._execute(variable=variable,
                                 context=context,
                                 runtime_params=runtime_params,
                                )
        if self._dictionary_memory:
            # Only return content if assoc has not been specified (in which case second element of value should be empty)
            if len(value[1]) == 0:
                return np.delete(value,1)
        return value

    @property
    def memory(self):
        """Return function's memory attribute"""
        try:
            return self.function.memory
        except:
            raise EpisodicMemoryMechanismError(f"Function of {self.name} ({self.function.name}) "
                                               f"has no `memory attribute")



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* LLVM IR Generation **************************************************************
import ast
import warnings
import numpy as np
from functools import reduce

from llvmlite import ir
from contextlib import contextmanager


from psyneulink.core.globals.keywords import AFTER, BEFORE
from psyneulink.core.scheduling.condition import Never
from psyneulink.core.scheduling.time import TimeScale
from . import helpers
from .debug import debug_env

class UserDefinedFunctionVisitor(ast.NodeVisitor):
    def __init__(self, ctx, builder, func_globals, func_params, arg_in, arg_out):
        self.ctx = ctx
        self.builder = builder
        self.func_params = func_params
        self.arg_in = arg_in
        self.arg_out = arg_out

        #setup default functions
        def _list_sum(x):
            # HACK: Obtain polymorphic addition function by visiting the add node
            # this should ideally be moved to an explicit helper
            add_func = self.visit_Add(None)

            total_sum = builder.alloca(x.type.pointee.element)
            builder.store(total_sum.type.pointee(None), total_sum)
            with helpers.array_ptr_loop(builder, x, "list_sum") as (b, idx):
                curr_val = b.gep(x, [ctx.int32_ty(0), idx])
                tmp = add_func(total_sum, curr_val)
                if helpers.is_pointer(tmp):
                    tmp = builder.load(tmp)
                b.store(tmp, total_sum)
            return total_sum

        def _len(x):
            x_ty = x.type
            if helpers.is_pointer(x):
                x_ty = x_ty.pointee
            return ctx.float_ty(len(x_ty))

        def _tanh(x):
            output_ptr = builder.alloca(x.type.pointee)
            helpers.call_elementwise_operation(self.ctx, self.builder, x, helpers.tanh, output_ptr)
            return output_ptr

        def _exp(x):
            output_ptr = builder.alloca(x.type.pointee)
            helpers.call_elementwise_operation(self.ctx, self.builder, x, helpers.exp, output_ptr)
            return output_ptr

        # numpy's max function differs greatly from that of python's buiiltin max
        # see: https://numpy.org/doc/stable/reference/generated/numpy.amax.html#numpy.amax
        def _max_numpy(x):
            assert helpers.is_vector(x) or helpers.is_2d_matrix(x), "Attempted to call max on invalid variable! Only 1-d and 2-d lists are supported!"
            curr = builder.alloca(ctx.float_ty)
            builder.store(ctx.float_ty('NaN'), curr)
            for (element_ptr,) in helpers.recursive_iterate_arrays(ctx, builder, x):
                element = builder.load(element_ptr)
                greater = builder.fcmp_unordered('>', element, builder.load(curr))
                with builder.if_then(greater):
                    builder.store(element, curr)
            return curr

        # see: https://docs.python.org/3/library/functions.html#max
        def _max(*args):
            if len(args) == 1 and helpers.is_vector(args[0]):
                curr = builder.alloca(ctx.float_ty)
                builder.store(ctx.float_ty('NaN'), curr)
                for (element_ptr,) in helpers.recursive_iterate_arrays(ctx, builder, args[0]):
                    element = builder.load(element_ptr)
                    greater = builder.fcmp_unordered('>', element, builder.load(curr))
                    with builder.if_then(greater):
                        builder.store(element, curr)
                return curr
            elif len(args) > 1 and all(a.type == args[0].type for a in args):
                curr = builder.alloca(ctx.float_ty)
                builder.store(ctx.float_ty('NaN'), curr)
                for element in args:
                    if helpers.is_pointer(element):
                        element = builder.load(element)
                    greater = builder.fcmp_unordered('>', element, builder.load(curr))
                    with builder.if_then(greater):
                        builder.store(element, curr)
                return curr
            assert False, "Attempted to call max with invalid arguments!"
        self.register = {
            "sum": _list_sum,
            "len": _len,
            "float": ctx.float_ty,
            "int": ctx.int32_ty,
            "max": _max,
        }

        # setup numpy
        numpy_handlers = {
            'tanh': _tanh,
            'exp': _exp,
            'equal': self._generate_fcmp_handler(self.ctx, self.builder, "=="),
            'not_equal': self._generate_fcmp_handler(self.ctx, self.builder, "!="),
            'less': self._generate_fcmp_handler(self.ctx, self.builder, "<"),
            'less_equal': self._generate_fcmp_handler(self.ctx, self.builder, "<="),
            'greater': self._generate_fcmp_handler(self.ctx, self.builder, ">"),
            'greater_equal': self._generate_fcmp_handler(self.ctx, self.builder, ">="),
            "max": _max_numpy,
        }

        for k, v in func_globals.items():
            if v is np:
                self.register[k] = numpy_handlers

        name_constants = {
            True: ctx.bool_ty(1),
            False: ctx.bool_ty(0),
        }
        self.name_constants = name_constants
        super().__init__()

    def visit_arguments(self, node):
        args = node.args
        variable = args[0]
        # update register
        self.register[variable.arg] = self.arg_in
        parameters = args[1:]
        for param in parameters:
            assert param.arg not in ["self", "owner"], f"Unable to reference {param.arg} in a compiled UserDefinedFunction!"
            if param.arg == 'params':
                assert False, "Runtime parameters are not supported in compiled mode"
            elif param.arg == 'context':
                # Since contexts are implicit in the structs in compiled mode, we do not compile it.
                pass
            else:
                self.register[param.arg] = self.func_params[param.arg]

    def _generate_binop(self, x, y, callback):
        # unpack scalars from pointers
        if helpers.is_floating_point(x) and helpers.is_pointer(x):
            x = self.builder.load(x)
        if helpers.is_floating_point(y) and helpers.is_pointer(y):
            y = self.builder.load(y)

        return callback(self.ctx, self.builder, x, y)

    def visit_Add(self, node):
        def _add_vec(ctx, builder, u, v):
            assert u.type == v.type

            dim = len(u.type.pointee)
            output_vec = builder.alloca(u.type.pointee)

            # Get the pointer to the first element of the array to convert from [? x double]* -> double*
            vec_u = builder.gep(u, [ctx.int32_ty(0), ctx.int32_ty(0)])
            vec_v = builder.gep(v, [ctx.int32_ty(0), ctx.int32_ty(0)])
            vec_out = builder.gep(output_vec, [ctx.int32_ty(0), ctx.int32_ty(0)])

            builtin = ctx.import_llvm_function("__pnl_builtin_vec_add")
            builder.call(builtin, [vec_u, vec_v, ctx.int32_ty(dim), vec_out])
            return output_vec

        def _add_mat(ctx, builder, m1, m2):
            assert m1.type == m2.type

            x = len(m1.type.pointee)
            y = len(m1.type.pointee.element)

            output_mat = builder.alloca(m1.type.pointee)

            m1_ptr = builder.gep(m1, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(0)])
            m2_ptr = builder.gep(m2, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(0)])
            output_ptr = builder.gep(output_mat, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(0)])

            builtin = ctx.import_llvm_function("__pnl_builtin_mat_add")
            builder.call(builtin, [m1_ptr, m2_ptr, ctx.int32_ty(x), ctx.int32_ty(y), output_ptr])
            return output_mat

        def _add_mat_scalar(ctx, builder, m1, s):
            x = len(m1.type.pointee)
            y = len(m1.type.pointee.element)
            output_mat = builder.alloca(m1.type.pointee)

            m1_ptr = builder.gep(m1, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(0)])
            output_ptr = builder.gep(output_mat, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(0)])

            builtin = ctx.import_llvm_function("__pnl_builtin_mat_scalar_add")
            builder.call(builtin, [m1_ptr, s, ctx.int32_ty(x), ctx.int32_ty(y), output_ptr])
            return output_mat

        def _add_vec_scalar(ctx, builder, u, s):
            output_vec = builder.alloca(u.type.pointee)

            helpers.call_elementwise_operation(ctx, builder, u, lambda ctx, builder, x: builder.fadd(x, s), output_vec)
            return output_vec

        def _add(x, y):
            if helpers.is_floating_point(x) and helpers.is_floating_point(y):
                return self._generate_binop(x, y, lambda ctx, builder, x, y: builder.fadd(x,y))
            elif helpers.is_vector(x) and helpers.is_floating_point(y):
                return self._generate_binop(x, y, _add_vec_scalar)
            elif helpers.is_floating_point(x) and helpers.is_vector(y):
                return self._generate_binop(y, x, _add_vec_scalar)
            elif helpers.is_2d_matrix(x) and helpers.is_floating_point(y):
                return self._generate_binop(x, y, _add_mat_scalar)
            elif helpers.is_floating_point(x) and helpers.is_2d_matrix(y):
                return self._generate_binop(y, x, _add_mat_scalar)
            elif helpers.is_vector(x) and helpers.is_vector(y):
                return self._generate_binop(x, y, _add_vec)
            elif helpers.is_2d_matrix(x) and helpers.is_2d_matrix(y):
                return self._generate_binop(x, y, _add_mat)

        return _add

    def visit_Mult(self, node):
        def _mul_vec(ctx, builder, u, v):
            assert u.type == v.type

            dim = len(u.type.pointee)
            output_vec = builder.alloca(u.type.pointee)

            # Get the pointer to the first element of the array to convert from [? x double]* -> double*
            vec_u = builder.gep(u, [ctx.int32_ty(0), ctx.int32_ty(0)])
            vec_v = builder.gep(v, [ctx.int32_ty(0), ctx.int32_ty(0)])
            vec_out = builder.gep(output_vec, [ctx.int32_ty(0), ctx.int32_ty(0)])

            builtin = ctx.import_llvm_function("__pnl_builtin_vec_hadamard")
            builder.call(builtin, [vec_u, vec_v, ctx.int32_ty(dim), vec_out])
            return output_vec

        def _mul_mat(ctx, builder, m1, m2):
            assert m1.type == m2.type

            x = len(m1.type.pointee)
            y = len(m1.type.pointee.element)

            output_mat = builder.alloca(m1.type.pointee)

            m1_ptr = builder.gep(m1, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(0)])
            m2_ptr = builder.gep(m2, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(0)])
            output_ptr = builder.gep(output_mat, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(0)])

            builtin = ctx.import_llvm_function("__pnl_builtin_mat_hadamard")
            builder.call(builtin, [m1_ptr, m2_ptr, ctx.int32_ty(x), ctx.int32_ty(y), output_ptr])
            return output_mat

        def _mul_mat_scalar(ctx, builder, m1, s):
            x = len(m1.type.pointee)
            y = len(m1.type.pointee.element)
            output_mat = builder.alloca(m1.type.pointee)

            m1_ptr = builder.gep(m1, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(0)])
            output_ptr = builder.gep(output_mat, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(0)])

            builtin = ctx.import_llvm_function("__pnl_builtin_mat_scalar_mult")
            builder.call(builtin, [m1_ptr, s, ctx.int32_ty(x), ctx.int32_ty(y), output_ptr])
            return output_mat

        def _mul_vec_scalar(ctx, builder, v, s):
            x = len(v.type.pointee)
            output_vec = builder.alloca(v.type.pointee)

            v_ptr = builder.gep(v, [ctx.int32_ty(0), ctx.int32_ty(0)])
            output_ptr = builder.gep(output_vec, [ctx.int32_ty(0), ctx.int32_ty(0)])

            builtin = ctx.import_llvm_function("__pnl_builtin_vec_scalar_mult")
            builder.call(builtin, [v_ptr, s, ctx.int32_ty(x), output_ptr])
            return output_vec

        def _mul(x, y):
            if helpers.is_floating_point(x) and helpers.is_floating_point(y):
                return self._generate_binop(x, y, lambda ctx, builder, x, y: builder.fmul(x,y))
            elif helpers.is_vector(x) and helpers.is_floating_point(y):
                return self._generate_binop(x, y, _mul_vec_scalar)
            elif helpers.is_floating_point(x) and helpers.is_vector(y):
                return self._generate_binop(y, x, _mul_vec_scalar)
            elif helpers.is_2d_matrix(x) and helpers.is_floating_point(y):
                return self._generate_binop(x, y, _mul_mat_scalar)
            elif helpers.is_floating_point(x) and helpers.is_2d_matrix(y):
                return self._generate_binop(y, x, _mul_mat_scalar)
            elif helpers.is_vector(x) and helpers.is_vector(y):
                if x.type != y.type:
                    # Special case: Cast y into scalar if it can be done
                    if helpers.get_array_shape(y) == [1]:
                        y = self.builder.gep(y, [self.ctx.int32_ty(0), self.ctx.int32_ty(0)])
                        return self._generate_binop(x, y, _mul_vec_scalar)
                return self._generate_binop(x, y, _mul_vec)
            elif helpers.is_2d_matrix(x) and helpers.is_2d_matrix(y):
                return self._generate_binop(x, y, _mul_mat)

        return _mul

    def visit_Div(self, node):
        def _div_array(ctx, builder, u, v):
            assert u.type == v.type
            output_ptr = builder.alloca(u.type.pointee)

            for (u_ptr, v_ptr, out_ptr) in helpers.recursive_iterate_arrays(ctx, builder, u, v, output_ptr):
                u_val = builder.load(u_ptr)
                v_val = builder.load(v_ptr)
                builder.store(builder.fdiv(u_val, v_val), out_ptr)

            return output_ptr

        def _div_array_scalar(ctx, builder, array, s):
            output_ptr = builder.alloca(array.type.pointee)
            helpers.call_elementwise_operation(ctx, builder, array, lambda ctx, builder, x:  builder.fdiv(x, s), output_ptr)

            return output_ptr

        def _div_scalar_array(ctx, builder, s, array):
            output_ptr = builder.alloca(array.type.pointee)
            helpers.call_elementwise_operation(ctx, builder, array, lambda ctx, builder, x:  builder.fdiv(s, x), output_ptr)

            return output_ptr

        def _div(x, y):
            if helpers.is_floating_point(x) and helpers.is_floating_point(y):
                return self._generate_binop(x, y, lambda ctx, builder, x, y: builder.fdiv(x, y))
            elif helpers.is_floating_point(x) and (helpers.is_2d_matrix(y) or helpers.is_vector(y)):
                return self._generate_binop(x, y, _div_scalar_array)
            elif helpers.is_vector(x) and helpers.is_floating_point(y):
                return self._generate_binop(x, y, _div_array_scalar)
            elif helpers.is_2d_matrix(x) and helpers.is_floating_point(y):
                return self._generate_binop(x, y, _div_array_scalar)
            elif helpers.is_vector(x) and helpers.is_vector(y):
                if x.type != y.type:
                    # Special case: Cast y into scalar if it can be done
                    if helpers.get_array_shape(y) == [1]:
                        y = self.builder.gep(y, [self.ctx.int32_ty(0), self.ctx.int32_ty(0)])
                        return self._generate_binop(x, y, _div_array_scalar)
                return self._generate_binop(x, y, _div_array)
            elif helpers.is_2d_matrix(x) and helpers.is_2d_matrix(y):
                return self._generate_binop(x, y, _div_array)
            assert False, f"Unable to divide arguments {x}, {y}"
        return _div

    def _generate_unop(self, x, callback):
        if helpers.is_floating_point(x) and helpers.is_pointer(x):
            x = self.builder.load(x)

        return callback(self.ctx, self.builder, x)

    def visit_USub(self, node):
        def _usub_vec_mat(ctx, builder, x):
            output = builder.alloca(x.type.pointee)

            helpers.call_elementwise_operation(ctx, builder, x, lambda ctx, builder, x: helpers.fneg(builder, x), output)
            return output

        def _usub(x):
            if helpers.is_floating_point(x):
                return self._generate_unop(x, lambda ctx, builder, x: helpers.fneg(builder, x))
            elif helpers.is_vector(x) or helpers.is_2d_matrix(x):
                return self._generate_unop(x, _usub_vec_mat)

        return _usub

    def visit_Name(self, node):
        return self.register[node.id]

    def visit_Attribute(self, node):
        val = self.visit(node.value)

        # special case numpy attributes
        if node.attr == "shape":
            shape = helpers.get_array_shape(val)
            return ir.ArrayType(self.ctx.float_ty, len(shape))(shape)
        elif node.attr == "flatten":
            def flatten():
                shape = helpers.get_array_shape(val)
                flattened_size = reduce(lambda x, y: x * y, shape)
                flattened_ty = ir.ArrayType(self.ctx.float_ty, flattened_size)
                flattened_array = self.builder.alloca(flattened_ty)
                index_var = self.builder.alloca(self.ctx.int32_ty, name="flattened_index_var_loc")
                self.builder.store(self.ctx.int32_ty(0), index_var)
                for (array_ptr,) in helpers.recursive_iterate_arrays(self.ctx, self.builder, val):
                    index = self.builder.load(index_var, name="flattened_index_var")
                    flattened_array_ptr = self.builder.gep(flattened_array, [self.ctx.int32_ty(0), index])
                    array_val = self.builder.load(array_ptr)
                    self.builder.store(array_val, flattened_array_ptr)
                    index = self.builder.add(index, self.ctx.int32_ty(1), name="flattened_index_var_inc")
                    self.builder.store(index, index_var)
                return flattened_array
            return flatten
        elif node.attr == "astype":
            def astype(ty):
                def _convert(ctx, builder, x):
                    if helpers.is_pointer(x):
                        x = builder.load(x)
                    if helpers.is_integer(x) and ty is ctx.float_ty:
                        if helpers.is_boolean(x):
                            return builder.uitofp(x, ty)
                        return builder.sitofp(x, ty)
                    elif helpers.is_floating_point(x) and ty is self.register["int"]:
                        return builder.fptosi(x, ty)
                    elif (helpers.is_floating_point(x) and ty is ctx.float_ty):
                        return x
                if helpers.is_scalar(val):
                    return _convert(self.ctx, self.builder, val)
                else:
                    output_ptr = self.builder.alloca(helpers.array_from_shape(helpers.get_array_shape(val), ty))
                    helpers.call_elementwise_operation(self.ctx, self.builder, val, _convert, output_ptr)
                    return output_ptr
            # we only support float types
            return astype

        return val[node.attr]

    def visit_Num(self, node):
        return self.ctx.float_ty(node.n)

    def visit_Assign(self, node):
        value = self.visit(node.value)
        def _assign_target(target, value):
            if isinstance(target, ast.Name):
                id = target.id
                self.register[id] = value
            else:
                to_store = value

                target = self.visit(target)
                self.builder.store(to_store, target)

        for target in node.targets:
            _assign_target(target, value)

    def visit_NameConstant(self, node):
        val = self.name_constants[node.value]
        assert val, f"Failed to convert NameConstant {node.value}"
        return val

    def visit_Tuple(self, node):
        elements = [self.visit(element) for element in node.elts]

        elements = [self.builder.load(element) if helpers.is_pointer(element) else element for element in elements]

        element_types = [element.type for element in elements]
        if len(element_types) > 0 and all(x == element_types[0] for x in element_types):
            ret_list = self.builder.alloca(ir.ArrayType(element_types[0], len(element_types)))
        else:
            ret_list = self.builder.alloca(ir.LiteralStructType(element_types))

        for idx, element in enumerate(elements):
            self.builder.store(element, self.builder.gep(ret_list, [self.ctx.int32_ty(0), self.ctx.int32_ty(idx)]))

        return ret_list

    def visit_BinOp(self, node):
        operator = self.visit(node.op)
        return operator(self.visit(node.left), self.visit(node.right))

    def visit_BoolOp(self, node):
        operator = self.visit(node.op)
        values = [self.visit(value) for value in node.values]
        ret_val = values[0]
        for value in values[1:]:
            ret_val = operator(ret_val, value)
        return ret_val

    def visit_UnaryOp(self, node):
        operator = self.visit(node.op)
        return operator(self.visit(node.operand))

    def visit_List(self, node):
        elements = [self.visit(element) for element in node.elts]
        element_ty = elements[0].type
        assert all(element.type == element_ty for element in elements), f"Unable to convert {node} into a list! (Elements differ in type!)"

        # dereference pointers
        if helpers.is_pointer(element_ty):
            elements = [self.builder.load(element) for element in elements]
            element_ty = elements[0].type

        ret_list = self.builder.alloca(ir.ArrayType(element_ty, len(elements)))

        for idx, element in enumerate(elements):
            self.builder.store(element, self.builder.gep(ret_list, [self.ctx.int32_ty(0), self.ctx.int32_ty(idx)]))
        return ret_list

    def visit_And(self, node):
        def _and(x, y):
            assert helpers.is_boolean(x), f"{x} is not a Boolean!"
            assert helpers.is_boolean(y), f"{y} is not a Boolean!"
            return self._generate_binop(x, y, lambda ctx, builder, x, y: builder.and_(x, y))

        return _and

    def visit_Or(self, node):
        def _or(x, y):
            assert helpers.is_boolean(x), f"{x} is not a Boolean!"
            assert helpers.is_boolean(y), f"{y} is not a Boolean!"
            return self._generate_binop(x, y, lambda ctx, builder, x, y: builder.or_(x, y))

        return _or

    def _generate_fcmp_handler(self, ctx, builder, cmp):
        def _cmp_array(ctx, builder, u, v):
            assert u.type == v.type
            shape = helpers.get_array_shape(u)
            output_ptr = builder.alloca(helpers.array_from_shape(shape, ctx.bool_ty))

            for (u_ptr, v_ptr, out_ptr) in helpers.recursive_iterate_arrays(ctx, builder, u, v, output_ptr):
                u_val = builder.load(u_ptr)
                v_val = builder.load(v_ptr)
                builder.store(builder.fcmp_ordered(cmp, u_val, v_val), out_ptr)

            return output_ptr

        def _cmp_array_scalar(ctx, builder, array, s):
            shape = helpers.get_array_shape(array)
            output_ptr = builder.alloca(helpers.array_from_shape(shape, ctx.bool_ty))
            helpers.call_elementwise_operation(ctx, builder, array, lambda ctx, builder, x:  builder.fcmp_ordered(cmp, x, s), output_ptr)

            return output_ptr

        def _cmp_scalar_array(ctx, builder, s, array):
            shape = helpers.get_array_shape(array)
            output_ptr = builder.alloca(helpers.array_from_shape(shape, ctx.bool_ty))
            helpers.call_elementwise_operation(ctx, builder, array, lambda ctx, builder, x:  builder.fcmp_ordered(cmp, s, x), output_ptr)

            return output_ptr

        def _cmp(x, y):
            if helpers.is_floating_point(x) and helpers.is_floating_point(y):
                return self._generate_binop(x, y, lambda ctx, builder, x, y: builder.fcmp_ordered(cmp, x, y))
            elif helpers.is_vector(x) and helpers.is_floating_point(y):
                return self._generate_binop(x, y, _cmp_array_scalar)
            elif helpers.is_floating_point(x) and helpers.is_vector(y):
                return self._generate_binop(x, y, _cmp_scalar_array)
            elif helpers.is_2d_matrix(x) and helpers.is_floating_point(y):
                return self._generate_binop(x, y, _cmp_array_scalar)
            elif helpers.is_floating_point(x) and helpers.is_2d_matrix(y):
                return self._generate_binop(x, y, _cmp_scalar_array)
            elif helpers.is_vector(x) and helpers.is_vector(y):
                return self._generate_binop(x, y, _cmp_array)
            elif helpers.is_2d_matrix(x) and helpers.is_2d_matrix(y):
                return self._generate_binop(x, y, _cmp_array)

        return _cmp

    def visit_Eq(self, node):
        return self._generate_fcmp_handler(self.ctx, self.builder, "==")

    def visit_NotEq(self, node):
        return self._generate_fcmp_handler(self.ctx, self.builder, "!=")

    def visit_Lt(self, node):
        return self._generate_fcmp_handler(self.ctx, self.builder, "<")

    def visit_LtE(self, node):
        return self._generate_fcmp_handler(self.ctx, self.builder, "<=")

    def visit_Gt(self, node):
        return self._generate_fcmp_handler(self.ctx, self.builder, ">")

    def visit_GtE(self, node):
        return self._generate_fcmp_handler(self.ctx, self.builder, ">=")

    def visit_Compare(self, node):
        comp_val = self.visit(node.left)
        comparators = [self.visit(comparator) for comparator in node.comparators]
        ops = [self.visit(op) for op in node.ops]
        for comparator, op in zip(comparators, ops):
            comp_val = op(comp_val, comparator)
        return comp_val

    def visit_If(self, node):
        predicate = self.visit(node.test)
        with self.builder.if_else(predicate) as (then, otherwise):
            with then:
                for child in node.body:
                    self.visit(child)
            with otherwise:
                for child in node.orelse:
                    self.visit(child)

    def visit_Return(self, node):
        ret_val = self.visit(node.value)
        arg_out = self.arg_out

        # dereference pointer
        if helpers.is_pointer(ret_val):
            ret_val = self.builder.load(ret_val)

        # get position in arg_out if types differ
        if (helpers.is_scalar(ret_val) or helpers.is_vector(ret_val)) and helpers.is_2d_matrix(arg_out):
            arg_out = self.builder.gep(arg_out, [self.ctx.int32_ty(0), self.ctx.int32_ty(0)])

        if helpers.is_scalar(ret_val) and helpers.is_vector(arg_out):
            arg_out = self.builder.gep(arg_out, [self.ctx.int32_ty(0), self.ctx.int32_ty(0)])

        self.builder.store(ret_val, arg_out)

    def visit_Call(self, node):
        call_func = self.visit(node.func)
        assert callable(call_func), f"Uncallable function {node.func}!"
        node_args = [self.visit(arg) for arg in node.args]

        return call_func(*node_args)

    def visit_Subscript(self, node):
        node_val = self.visit(node.value)
        node_slice_val = self.visit(node.slice)
        return self.builder.gep(node_val, [self.ctx.int32_ty(0), node_slice_val])

    def visit_Index(self, node):
        return self.builder.fptoui(self.visit(node.value), self.ctx.int32_ty)

def gen_node_wrapper(ctx, composition, node, *, tags:frozenset):
    assert "node_wrapper" in tags
    func_tags = tags.difference({"node_wrapper"})

    node_function = ctx.import_llvm_function(node, tags=func_tags)
    # FIXME: This is a hack
    is_mech = hasattr(node, 'function')

    data_struct_ptr = ctx.get_data_struct_type(composition).as_pointer()
    args = [
        ctx.get_state_struct_type(composition).as_pointer(),
        ctx.get_param_struct_type(composition).as_pointer(),
        ctx.get_input_struct_type(composition).as_pointer(),
        data_struct_ptr, data_struct_ptr]

    if not is_mech and "reset" not in tags:
        # Add condition struct of the parent composition
        # This includes structures of all nested compositions
        cond_gen = helpers.ConditionGenerator(ctx, composition)
        cond_ty = cond_gen.get_condition_struct_type().as_pointer()
        args.append(cond_ty)

    builder = ctx.create_llvm_function(args, node, tags=tags,
                                       return_type=node_function.type.pointee.return_type)
    llvm_func = builder.function
    for a in llvm_func.args:
        a.attributes.add('nonnull')

    state, params, comp_in, data_in, data_out = llvm_func.args[:5]

    if node is composition.input_CIM:
        # if there are incoming modulatory projections,
        # the input structure is shared
        if composition.parameter_CIM.afferents:
            node_in = builder.gep(comp_in, [ctx.int32_ty(0), ctx.int32_ty(0)])
        else:
            node_in = comp_in
        incoming_projections = []
    elif node is composition.parameter_CIM and node.afferents:
        # if parameter_CIM has afferent projections,
        # their values are in comp_in[1]
        node_in = builder.gep(comp_in, [ctx.int32_ty(0), ctx.int32_ty(1)])
        # And we run no further projection
        incoming_projections = []
    elif not is_mech:
        node_in = builder.alloca(node_function.args[2].type.pointee)
        incoming_projections = node.input_CIM.afferents + node.parameter_CIM.afferents
    else:
        # this path also handles parameter_CIM with no afferent
        # projections. 'comp_in' does not include any extra values,
        # and the entire call should be optimized out.
        node_in = builder.alloca(node_function.args[2].type.pointee)
        incoming_projections = node.afferents

    if "reset" in tags or "is_finished" in tags:
        incoming_projections = []

    # Execute all incoming projections
    inner_projections = list(composition._inner_projections)
    zero = ctx.int32_ty(0)
    projections_params = helpers.get_param_ptr(builder, composition,
                                               params, "projections")
    projections_states = helpers.get_state_ptr(builder, composition,
                                               state, "projections")
    for proj in incoming_projections:
        # Skip autoassociative projections.
        # Recurrent projections are executed as part of the mechanism to
        # make sure their value is up-to-date inside the 'is_finished' loop
        if proj.sender.owner is proj.receiver.owner:
            continue

        # Get location of projection input data
        par_mech = proj.sender.owner
        if par_mech in composition._all_nodes:
            parent_idx = composition._get_node_index(par_mech)
        else:
            assert par_mech is par_mech.composition.output_CIM
            parent_idx = composition.nodes.index(par_mech.composition)

        assert proj.sender in par_mech.output_ports
        output_port_idx = par_mech.output_ports.index(proj.sender)
        proj_in = builder.gep(data_in, [ctx.int32_ty(0),
                                        ctx.int32_ty(0),
                                        ctx.int32_ty(parent_idx),
                                        ctx.int32_ty(output_port_idx)])

        # Get location of projection output (in mechanism's input structure
        rec_port = proj.receiver
        assert rec_port.owner is node or rec_port.owner is node.input_CIM or rec_port.owner is node.parameter_CIM
        indices = [0]
        if proj in rec_port.owner.path_afferents:
            rec_port_idx = rec_port.owner.input_ports.index(rec_port)

            assert proj in rec_port.pathway_projections
            projection_idx = rec_port.pathway_projections.index(proj)

            if not is_mech and node.parameter_CIM.afferents:
                # If there are afferent projections to parameter_CIM
                # the input structure is split between input_CIM
                # and parameter_CIM
                if proj in node.parameter_CIM.afferents:
                    # modulatory projection
                    indices.append(1)
                else:
                    # pathway projection
                    indices.append(0)
            indices.extend([rec_port_idx, projection_idx])
        elif proj in rec_port.owner.mod_afferents:
            # Only mechanism ports list mod projections in mod_afferents
            assert is_mech
            projection_idx = rec_port.owner.mod_afferents.index(proj)
            indices.extend([len(rec_port.owner.input_ports), projection_idx])
        else:
            assert False, "Projection neither pathway nor modulatory"

        proj_out = builder.gep(node_in, [ctx.int32_ty(i) for i in indices])

        # Get projection parameters and state
        proj_idx = inner_projections.index(proj)
        proj_params = builder.gep(projections_params, [zero, ctx.int32_ty(proj_idx)])
        proj_state = builder.gep(projections_states, [zero, ctx.int32_ty(proj_idx)])
        proj_function = ctx.import_llvm_function(proj, tags=func_tags)

        if proj_out.type != proj_function.args[3].type:
            warnings.warn("Shape mismatch: Projection ({}) results does not match the receiver state({}) input: {} vs. {}".format(proj, proj.receiver, proj.defaults.value, proj.receiver.defaults.variable))
            proj_out = builder.bitcast(proj_out, proj_function.args[3].type)
        builder.call(proj_function, [proj_params, proj_state, proj_in, proj_out])


    node_idx = ctx.int32_ty(composition._get_node_index(node))
    nodes_params = helpers.get_param_ptr(builder, composition, params, "nodes")
    nodes_states = helpers.get_state_ptr(builder, composition, state, "nodes")
    node_params = builder.gep(nodes_params, [zero, node_idx])
    node_state = builder.gep(nodes_states, [zero, node_idx])
    node_out = builder.gep(data_out, [zero, zero, node_idx])
    if is_mech:
        call_args = [node_params, node_state, node_in, node_out]
        if len(node_function.args) > 4:
            assert node is composition.controller
            call_args += [params, state, data_in]
        ret = builder.call(node_function, call_args)
    elif "reset" not in tags:
        # FIXME: reinitialization of compositions is not supported
        # Condition and data structures includes parent first
        nested_idx = ctx.int32_ty(composition._get_node_index(node) + 1)
        node_data = builder.gep(data_in, [zero, nested_idx])
        node_cond = builder.gep(llvm_func.args[5], [zero, nested_idx])
        ret = builder.call(node_function, [node_state, node_params, node_in,
                                           node_data, node_cond])
        # Copy output of the nested composition to its output place
        output_idx = node._get_node_index(node.output_CIM)
        result = builder.gep(node_data, [zero, zero, ctx.int32_ty(output_idx)])
        builder.store(builder.load(result), node_out)
    else:
        # composition reset
        ret = None

    if ret is None or isinstance(ret.type, ir.VoidType):
        builder.ret_void()
    else:
        builder.ret(ret)

    return llvm_func


@contextmanager
def _gen_composition_exec_context(ctx, composition, *, tags:frozenset, suffix="", extra_args=[]):
    cond_gen = helpers.ConditionGenerator(ctx, composition)

    name = "_".join(("wrap_exec", *tags ,composition.name + suffix))
    args = [ctx.get_state_struct_type(composition).as_pointer(),
            ctx.get_param_struct_type(composition).as_pointer(),
            ctx.get_input_struct_type(composition).as_pointer(),
            ctx.get_data_struct_type(composition).as_pointer(),
            cond_gen.get_condition_struct_type().as_pointer()]
    builder = ctx.create_llvm_function(args + extra_args, composition, name)
    llvm_func = builder.function

    for a in llvm_func.args:
        a.attributes.add('noalias')

    state, params, comp_in, data_arg, cond, *_ = llvm_func.args
    if "const_params" in debug_env:
        const_params = params.type.pointee(composition._get_param_initializer(None))
        params = builder.alloca(const_params.type, name="const_params_loc")
        builder.store(const_params, params)

    if "alloca_data" in debug_env:
        data = builder.alloca(data_arg.type.pointee)
        data_vals = builder.load(data_arg)
        builder.store(data_vals, data)
    else:
        data = data_arg

    node_tags = tags.union({"node_wrapper"})
    # Call input CIM
    input_cim_w = ctx.get_node_wrapper(composition, composition.input_CIM)
    input_cim_f = ctx.import_llvm_function(input_cim_w, tags=node_tags)
    builder.call(input_cim_f, [state, params, comp_in, data, data])

    # Call parameter CIM
    param_cim_w = ctx.get_node_wrapper(composition, composition.parameter_CIM)
    param_cim_f = ctx.import_llvm_function(param_cim_w, tags=node_tags)
    builder.call(param_cim_f, [state, params, comp_in, data, data])

    yield builder, data, params, cond_gen

    if "alloca_data" in debug_env:
        data_vals = builder.load(data)
        builder.store(data_vals, data_arg)

    # Bump run counter
    cond_gen.bump_ts(builder, cond, (1, 0, 0))

    builder.ret_void()


def gen_composition_exec(ctx, composition, *, tags:frozenset):
    simulation = "simulation" in tags
    node_tags = tags.union({"node_wrapper"})

    with _gen_composition_exec_context(ctx, composition, tags=tags) as (builder, data, params, cond_gen):
        state, _, comp_in, _, cond = builder.function.args

        nodes_states = helpers.get_param_ptr(builder, composition, state, "nodes")

        # Allocate temporary output storage
        output_storage = builder.alloca(data.type.pointee, name="output_storage")

        # Get locations of number of executions.
        num_exec_locs = {}
        for idx, node in enumerate(composition._all_nodes):
            node_state = builder.gep(nodes_states, [ctx.int32_ty(0),
                                                    ctx.int32_ty(idx)])
            num_exec_locs[node] = helpers.get_state_ptr(builder, node,
                                                        node_state,
                                                        "num_executions")

        # Generate pointers to 'is_finished' callbacks
        is_finished_callbacks = {}
        for node in composition.nodes:
            args = [state, params, comp_in, data, output_storage]
            wrapper = ctx.get_node_wrapper(composition, node)
            is_finished_callbacks[node] = (wrapper, args)


        # Reset internal TRIAL/PASS/TIME_STEP clock for each node
        for time_loc in num_exec_locs.values():
            for scale in (TimeScale.TRIAL, TimeScale.PASS, TimeScale.TIME_STEP):
                num_exec_time_ptr = builder.gep(time_loc, [ctx.int32_ty(0),
                                                           ctx.int32_ty(scale.value)])
                builder.store(num_exec_time_ptr.type.pointee(0), num_exec_time_ptr)

        # Check if there's anything to reset
        for node in composition._all_nodes:
            when = getattr(node, "reset_stateful_function_when", Never())
            # FIXME: This should not be necessary. The code gets DCE'd,
            # but there are still some problems with generation
            # 'reset' function
            if node is composition.controller:
                continue

            reinit_cond = cond_gen.generate_sched_condition(
                builder, when, cond, node, is_finished_callbacks, num_exec_locs)
            with builder.if_then(reinit_cond):
                node_w = ctx.get_node_wrapper(composition, node)
                node_reinit_f = ctx.import_llvm_function(node_w, tags=node_tags.union({"reset"}))
                builder.call(node_reinit_f, [state, params, comp_in, data, data])

        # Run controller if it's enabled in 'BEFORE' mode
        if simulation is False and composition.enable_controller and \
           composition.controller_mode == BEFORE:
            assert composition.controller is not None
            controller_w = ctx.get_node_wrapper(composition, composition.controller)
            controller_f = ctx.import_llvm_function(controller_w,
                                                     tags=node_tags)
            builder.call(controller_f, [state, params, comp_in, data, data])


        # Allocate run set structure
        run_set_type = ir.ArrayType(ctx.bool_ty, len(composition.nodes))
        run_set_ptr = builder.alloca(run_set_type, name="run_set")
        builder.store(run_set_type(None), run_set_ptr)


        iter_ptr = builder.alloca(ctx.int32_ty, name="iter_counter")
        builder.store(ctx.int32_ty(0), iter_ptr)

        # Start the main loop structure
        loop_condition = builder.append_basic_block(name="scheduling_loop_condition")
        builder.branch(loop_condition)

        # Generate a while not 'end condition' loop
        builder.position_at_end(loop_condition)

        trial_term_cond = cond_gen.generate_sched_condition(
            builder, composition.termination_processing[TimeScale.TRIAL],
            cond, None, is_finished_callbacks, num_exec_locs)
        trial_cond = builder.not_(trial_term_cond, name="not_trial_term_cond")

        loop_body = builder.append_basic_block(name="scheduling_loop_body")
        exit_block = builder.append_basic_block(name="trial_exit")
        builder.cbranch(trial_cond, loop_body, exit_block)

        # Generate loop body
        builder.position_at_end(loop_body)

        zero = ctx.int32_ty(0)
        any_cond = ctx.bool_ty(0)
        # Calculate execution set before running the mechanisms
        for idx, node in enumerate(composition.nodes):
            run_set_node_ptr = builder.gep(run_set_ptr,
                                           [zero, ctx.int32_ty(idx)],
                                           name="run_cond_ptr_" + node.name)
            node_cond = cond_gen.generate_sched_condition(
                builder, composition._get_processing_condition_set(node),
                cond, node, is_finished_callbacks, num_exec_locs)
            ran = cond_gen.generate_ran_this_pass(builder, cond, node)
            node_cond = builder.and_(node_cond, builder.not_(ran),
                                     name="run_cond_" + node.name)
            any_cond = builder.or_(any_cond, node_cond, name="any_ran_cond")
            builder.store(node_cond, run_set_node_ptr)

        # Reset internal TIME_STEP clock for each node
        # NOTE: This is done _after_ condition evluation, otherwise
        #       TIME_STEP related conditions will only see 0 executions
        for time_loc in num_exec_locs.values():
            num_exec_time_ptr = builder.gep(time_loc, [ctx.int32_ty(0),
                                                       ctx.int32_ty(TimeScale.TIME_STEP.value)])
            builder.store(num_exec_time_ptr.type.pointee(0), num_exec_time_ptr)

        for idx, node in enumerate(composition.nodes):

            run_set_node_ptr = builder.gep(run_set_ptr, [zero, ctx.int32_ty(idx)])
            node_cond = builder.load(run_set_node_ptr,
                                     name="node_" + node.name + "_should_run")
            with builder.if_then(node_cond):
                node_w = ctx.get_node_wrapper(composition, node)
                node_f = ctx.import_llvm_function(node_w, tags=node_tags)
                builder.block.name = "invoke_" + node_f.name
                # Wrappers do proper indexing of all structures
                # Mechanisms have only 5 args
                args = [state, params, comp_in, data, output_storage]
                if len(node_f.args) >= 6:  # Composition wrappers have 6 args
                    args.append(cond)
                builder.call(node_f, args)

                cond_gen.generate_update_after_run(builder, cond, node)
            builder.block.name = "post_invoke_" + node_f.name

        # Writeback results
        for idx, node in enumerate(composition.nodes):
            run_set_node_ptr = builder.gep(run_set_ptr, [zero, ctx.int32_ty(idx)])
            node_cond = builder.load(run_set_node_ptr, name="node_" + node.name + "_ran")
            with builder.if_then(node_cond):
                out_ptr = builder.gep(output_storage, [zero, zero,
                                                       ctx.int32_ty(idx)],
                                      name="result_ptr_" + node.name)
                data_ptr = builder.gep(data, [zero, zero, ctx.int32_ty(idx)],
                                       name="data_result_" + node.name)
                builder.store(builder.load(out_ptr), data_ptr)

        # Update step counter
        # FIXME: This shouldn't really happen. If no node is run,
        #        there's no change to the scheduling conditions and we hang.
        with builder.if_then(any_cond):
            builder.block.name = "inc_step"
            cond_gen.bump_ts(builder, cond)

        builder.block.name = "update_iter_count"
        # Increment number of iterations
        iters = builder.load(iter_ptr, name="iterw")
        iters = builder.add(iters, ctx.int32_ty(1), name="iterw_inc")
        builder.store(iters, iter_ptr)

        max_iters = len(composition.scheduler.consideration_queue)
        completed_pass = builder.icmp_unsigned("==", iters,
                                               ctx.int32_ty(max_iters),
                                               name="completed_pass")
        # Increment pass and reset time step
        with builder.if_then(completed_pass):
            builder.block.name = "inc_pass"
            builder.store(zero, iter_ptr)
            # Bumping automatically zeros lower elements
            cond_gen.bump_ts(builder, cond, (0, 1, 0))
            # Reset internal PASS clock for each node
            for time_loc in num_exec_locs.values():
                num_exec_time_ptr = builder.gep(time_loc, [ctx.int32_ty(0),
                                                           ctx.int32_ty(TimeScale.PASS.value)])
                builder.store(num_exec_time_ptr.type.pointee(0), num_exec_time_ptr)

        builder.branch(loop_condition)

        builder.position_at_end(exit_block)

        if simulation is False and composition.enable_controller and \
           composition.controller_mode == AFTER:
            assert composition.controller is not None
            controller_w = ctx.get_node_wrapper(composition, composition.controller)
            controller_f = ctx.import_llvm_function(controller_w, tags=node_tags)
            builder.call(controller_f, [state, params, comp_in, data, data])

        # Call output CIM
        output_cim_w = ctx.get_node_wrapper(composition, composition.output_CIM)
        output_cim_f = ctx.import_llvm_function(output_cim_w, tags=node_tags)
        builder.block.name = "invoke_" + output_cim_f.name
        builder.call(output_cim_f, [state, params, comp_in, data, data])

    return builder.function


def gen_composition_run(ctx, composition, *, tags:frozenset):
    assert "run" in tags
    simulation = "simulation" in tags
    name = "_".join(("wrap",  *tags, composition.name))
    args = [ctx.get_state_struct_type(composition).as_pointer(),
            ctx.get_param_struct_type(composition).as_pointer(),
            ctx.get_data_struct_type(composition).as_pointer(),
            ctx.get_input_struct_type(composition).as_pointer(),
            ctx.get_output_struct_type(composition).as_pointer(),
            ctx.int32_ty.as_pointer(),
            ctx.int32_ty.as_pointer()]
    builder = ctx.create_llvm_function(args, composition, name)
    llvm_func = builder.function
    for a in llvm_func.args:
        a.attributes.add('noalias')

    state, params, data, data_in, data_out, trials_ptr, inputs_ptr = llvm_func.args
    # simulation does not care about the output
    # it extracts results of the controller objective mechanism
    if simulation:
        data_out.attributes.remove('nonnull')

    if not simulation and "const_data" in debug_env:
        const_data = data.type.pointee(composition._get_data_initializer(None))
        data = builder.alloca(data.type.pointee)
        builder.store(const_data, data)

    # Hardcode stateful parameters if set in the environment
    if not simulation and "const_state" in debug_env:
        const_state = state.type.pointee(composition._get_state_initializer(None))
        state = builder.alloca(const_state.type, name="const_state_loc")
        builder.store(const_state, state)

    if not simulation and "const_input" in debug_env:
        if not debug_env["const_input"]:
            input_init = [[os.defaults.variable.tolist()] for os in composition.input_CIM.input_ports]
            print("Setting default input: ", input_init)
        else:
            input_init = ast.literal_eval(debug_env["const_input"])
            print("Setting user input: ", input_init)

        builder.store(data_in.type.pointee(input_init), data_in)
        builder.store(inputs_ptr.type.pointee(1), inputs_ptr)

    # Allocate and initialize condition structure
    cond_gen = helpers.ConditionGenerator(ctx, composition)
    cond_type = cond_gen.get_condition_struct_type()
    cond = builder.alloca(cond_type)
    cond_init = cond_type(cond_gen.get_condition_initializer())
    builder.store(cond_init, cond)

    trials = builder.load(trials_ptr, "trials")
    iters_ptr = builder.alloca(trials.type)
    builder.store(iters_ptr.type.pointee(0), iters_ptr)

    # Start the main loop structure
    loop_condition = builder.append_basic_block(name="run_loop_condition")
    builder.branch(loop_condition)

    # Generate a while not 'end condition' loop
    builder.position_at_end(loop_condition)

    run_term_cond = cond_gen.generate_sched_condition(
        builder, composition.termination_processing[TimeScale.RUN],
        cond, None, None, None)
    run_cond = builder.not_(run_term_cond, name="not_run_term_cond")

    # Iter cond
    iters = builder.load(iters_ptr)
    iter_cond = builder.icmp_unsigned("<", iters, trials)

    # Increment. Use new name to not taint 'iters'
    new_iters = builder.add(iters, iters.type(1))
    builder.store(new_iters, iters_ptr)

    loop_body = builder.append_basic_block(name="run_loop_body")
    exit_block = builder.append_basic_block(name="run_exit")

    run_cond = builder.and_(run_cond, iter_cond)
    builder.cbranch(run_cond, loop_body, exit_block)

    # Generate loop body
    builder.position_at_end(loop_body)

    # Get the right input stimulus
    input_idx = builder.urem(iters, builder.load(inputs_ptr))
    data_in_ptr = builder.gep(data_in, [input_idx])

    # Reset internal 'RUN' clocks of each node
    for idx, node in enumerate(composition._all_nodes):
        node_state = builder.gep(state, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(idx)])
        num_executions_ptr = helpers.get_state_ptr(builder, node, node_state, "num_executions")
        num_exec_time_ptr = builder.gep(num_executions_ptr, [ctx.int32_ty(0), ctx.int32_ty(TimeScale.RUN.value)])
        builder.store(num_exec_time_ptr.type.pointee(0), num_exec_time_ptr)

    # Call execution
    exec_tags = tags.difference({"run"})
    exec_f = ctx.import_llvm_function(composition, tags=exec_tags)
    builder.call(exec_f, [state, params, data_in_ptr, data, cond])

    if not simulation:
        # Extract output_CIM result
        idx = composition._get_node_index(composition.output_CIM)
        result_ptr = builder.gep(data, [ctx.int32_ty(0), ctx.int32_ty(0),
                                        ctx.int32_ty(idx)])
        output_ptr = builder.gep(data_out, [iters])
        result = builder.load(result_ptr)
        builder.store(result, output_ptr)

    builder.branch(loop_condition)

    # Exit
    builder.position_at_end(exit_block)
    # Store number of executed trials
    builder.store(iters, trials_ptr)

    builder.ret_void()
    return llvm_func


def gen_multirun_wrapper(ctx, function: ir.Function) -> ir.Function:
    if function.module is not ctx.module:
        function = ir.Function(ctx.module, function.type.pointee, function.name)
        assert function.is_declaration

    args = [a.type for a in function.args]
    args.append(ctx.int32_ty.as_pointer())
    multirun_ty = ir.FunctionType(function.type.pointee.return_type, args)
    multirun_f = ir.Function(ctx.module, multirun_ty, function.name + "_multirun")
    block = multirun_f.append_basic_block(name="entry")
    builder = ir.IRBuilder(block)

    multi_runs = builder.load(multirun_f.args[-1])
    # Runs need special handling. data_in and data_out are one dimensional,
    # but hold entries for all parallel invocations.
    is_comp_run = len(function.args) == 7
    if is_comp_run:
        trials_count = builder.load(multirun_f.args[5])
        input_count = builder.load(multirun_f.args[6])

    with helpers.for_loop_zero_inc(builder, multi_runs, "multi_run_loop") as (b, index):
        # Index all pointer arguments
        indexed_args = []
        for i, arg in enumerate(multirun_f.args[:-1]):
            # Don't adjust #inputs and #trials
            if isinstance(arg.type, ir.PointerType):
                offset = index
                # #runs and #trials needs to be the same for every invocation
                if is_comp_run and i >= 5:
                    offset = ctx.int32_ty(0)
                    # Reset trial count for every invocation.
                    # Previous runs might have finished earlier
                    if i == 5:
                        builder.store(trials_count, arg)
                # data arrays need special handling
                elif is_comp_run and i == 4:  # data_out
                    offset = b.mul(index, trials_count)
                elif is_comp_run and i == 3:  # data_in
                    offset = b.mul(index, input_count)

                arg = b.gep(arg, [offset])

            indexed_args.append(arg)

        b.call(function, indexed_args)

    builder.ret_void()
    return multirun_f


def gen_autodiffcomp_exec(ctx, composition, *, tags:frozenset):
    """Creates llvm bin execute for autodiffcomp"""
    assert composition.controller is None
    composition._build_pytorch_representation(composition.default_execution_id)
    pytorch_model = composition.parameters.pytorch_representation.get(composition.default_execution_id)
    with _gen_composition_exec_context(ctx, composition, tags=tags) as (builder, data, params, cond_gen):
        state, _, comp_in, _, cond = builder.function.args

        pytorch_func = ctx.import_llvm_function(pytorch_model, tags=tags)
        builder.call(pytorch_func, [state, params, data])

        node_tags = tags.union({"node_wrapper"})
        # Call output CIM
        output_cim_w = ctx.get_node_wrapper(composition, composition.output_CIM)
        output_cim_f = ctx.import_llvm_function(output_cim_w, tags=node_tags)
        builder.call(output_cim_f, [state, params, comp_in, data, data])

        return builder.function



import numpy as np
import pytest

import psyneulink.core.components.functions.transferfunctions as Functions
import psyneulink.core.llvm as pnlvm

@pytest.mark.function
@pytest.mark.identity_function
@pytest.mark.benchmark(group="IdentityFunction")
@pytest.mark.parametrize("size", [1, 2, 4, 8, 16])
def test_basic(size, benchmark, func_mode):
    variable = np.random.rand(size)
    f = Functions.Identity(default_variable=variable)
    if func_mode == 'Python':
        EX = f.function
    elif func_mode == 'LLVM':
        EX = pnlvm.execution.FuncExecution(f).execute
    elif func_mode == 'PTX':
        EX = pnlvm.execution.FuncExecution(f).cuda_execute

    res = benchmark(EX, variable)
    assert np.allclose(res, variable)



import numpy as np
import pytest

import psyneulink.core.llvm as pnlvm
from psyneulink.core.components.functions.function import FunctionError
from psyneulink.core.components.functions.statefulfunctions.memoryfunctions import DictionaryMemory, \
    ContentAddressableMemory
from psyneulink.library.components.mechanisms.processing.integrator.episodicmemorymechanism import \
    EpisodicMemoryMechanism

np.random.seed(0)

# TEST WITH DictionaryMemory ****************************************************************************************

CONTENT_SIZE=10
ASSOC_SIZE=10
test_var = np.random.rand(2, CONTENT_SIZE)
test_initializer = {tuple(test_var[0]): test_var[1]}

test_data = [
    (test_var, DictionaryMemory, {'default_variable':test_var}, [[
       0.5488135039273248, 0.7151893663724195, 0.6027633760716439, 0.5448831829968969, 0.4236547993389047,
        0.6458941130666561, 0.4375872112626925, 0.8917730007820798, 0.9636627605010293, 0.3834415188257777],
        [0.7917250380826646, 0.5288949197529045, 0.5680445610939323, 0.925596638292661, 0.07103605819788694,
         0.08712929970154071, 0.02021839744032572, 0.832619845547938, 0.7781567509498505, 0.8700121482468192 ]]),
    (test_var, DictionaryMemory, {'default_variable':test_var, 'retrieval_prob':0.5},
     [[ 0. for i in range(CONTENT_SIZE) ],[ 0. for i in range(ASSOC_SIZE) ]]),
    (test_var, DictionaryMemory, {'default_variable':test_var, 'storage_prob':0.1},
     [[ 0. for i in range(CONTENT_SIZE) ],[ 0. for i in range(ASSOC_SIZE) ]]),
    (test_var, DictionaryMemory, {'default_variable':test_var, 'retrieval_prob':0.9, 'storage_prob':0.9}, [[
       0.5488135039273248, 0.7151893663724195, 0.6027633760716439, 0.5448831829968969, 0.4236547993389047,
        0.6458941130666561, 0.4375872112626925, 0.8917730007820798, 0.9636627605010293, 0.3834415188257777],
        [0.7917250380826646, 0.5288949197529045, 0.5680445610939323, 0.925596638292661, 0.07103605819788694,
         0.08712929970154071, 0.02021839744032572, 0.832619845547938, 0.7781567509498505, 0.8700121482468192 ]]),
]

# use list, naming function produces ugly names
names = [
    "DictionaryMemory",
    "DictionaryMemory Random Retrieval",
    "DictionaryMemory Random Storage",
    "DictionaryMemory Random Retrieval-Storage",
]

@pytest.mark.function
@pytest.mark.memory_function
@pytest.mark.benchmark
@pytest.mark.parametrize('variable, func, params, expected', test_data, ids=names)
def test_with_dictionary_memory(variable, func, params, expected, benchmark, mech_mode):
    f = func(seed=0, **params)
    m = EpisodicMemoryMechanism(content_size=len(variable[0]), assoc_size=len(variable[1]), function=f)
    if mech_mode == 'Python':
        def EX(variable):
            m.execute(variable)
            return m.output_values
    elif mech_mode == 'LLVM':
        EX = pnlvm.execution.MechExecution(m).execute
    elif mech_mode == 'PTX':
        EX = pnlvm.execution.MechExecution(m).cuda_execute

    EX(variable)
    res = EX(variable)
    assert np.allclose(res[0], expected[0])
    assert np.allclose(res[1], expected[1])
    if benchmark.enabled:
        benchmark(EX, variable)

# TEST WITH ContentAddressableMemory ***********************************************************************************
# Note:  ContentAddressableMemory has not yet been compiled for use with LLVM or PTX, so those are dummy tests for now

test_data = [
    (
        # name
        "ContentAddressableMemory Func Default Variable Mech Size Init",
        # func
        ContentAddressableMemory,
        # func_params
        {'default_variable': [[0],[0,0],[0,0,0]]},
        # mech_params
        {'size':[1,2,3]},
        # test_var
        [[10.],[20., 30.],[40., 50., 60.]],
        # expected input_port names
        ['FIELD_0_INPUT', 'FIELD_1_INPUT', 'FIELD_2_INPUT'],
        # expected output_port names
        ['RETREIVED_FIELD_0', 'RETREIVED_FIELD_1', 'RETREIVED_FIELD_2'],
        # expected output
        [[10.],[20., 30.],[40., 50., 60.]]
    ),
    (
        "ContentAddressableMemory Func Default Variable Mech Default Var Init",
        ContentAddressableMemory,
        {'default_variable': [[0],[0,0],[0,0,0]]},
        {'default_variable': [[0],[0,0],[0,0,0]]},
        [[10.],[20., 30.],[40., 50., 60.]],
        ['FIELD_0_INPUT', 'FIELD_1_INPUT', 'FIELD_2_INPUT'],
        ['RETREIVED_FIELD_0', 'RETREIVED_FIELD_1', 'RETREIVED_FIELD_2'],
        [[10.],[20., 30.],[40., 50., 60.]]
    ),
    (
        "ContentAddressableMemory Func Initializer (ragged) Mech Size Init",
        ContentAddressableMemory,
        {'initializer':np.array([[np.array([1]), np.array([2, 3]), np.array([4, 5, 6])],
                                 [list([10]), list([20, 30]), list([40, 50, 60])],
                                 [np.array([11]), np.array([22, 33]), np.array([44, 55, 66])]], dtype=object)},
        {'size':[1,2,3]},
        [[10.],[20., 30.],[40., 50., 60.]],
        ['FIELD_0_INPUT', 'FIELD_1_INPUT', 'FIELD_2_INPUT'],
        ['RETREIVED_FIELD_0', 'RETREIVED_FIELD_1', 'RETREIVED_FIELD_2'],
        [[10.],[20., 30.],[40., 50., 60.]]
    ),
    (
        "ContentAddressableMemory Func Initializer (ragged) Mech Default Variable Init",
        ContentAddressableMemory,
        {'initializer':np.array([[np.array([1]), np.array([2, 3]), np.array([4, 5, 6])],
                                 [[10], [20, 30], [40, 50, 60]],
                                 [np.array([11]), np.array([22, 33]), np.array([44, 55, 66])]], dtype=object)},
        {'default_variable': [[0],[0,0],[0,0,0]], 'input_ports':['hello','world','goodbye']},
        [[10.],[20., 30.],[40., 50., 60.]],
        ['hello', 'world', 'goodbye'],
        ['RETREIVED_hello', 'RETREIVED_world', 'RETREIVED_goodbye'],
        [[10.],[20., 30.],[40., 50., 60.]]
    ),
    (
        "ContentAddressableMemory Func Initializer (regular 2d) Mech Size Init",
        ContentAddressableMemory,
        {'initializer':np.array([[np.array([1,2]), np.array([3,4]), np.array([5, 6])],
                                 [[10,20], [30,40], [50,60]],
                                 [np.array([11,12]), np.array([22, 23]), np.array([34, 35])]])},
        {'size':[2,2,2]},
        [[10,20], [30,40], [50, 60]],
        ['FIELD_0_INPUT', 'FIELD_1_INPUT', 'FIELD_2_INPUT'],
        ['RETREIVED_FIELD_0', 'RETREIVED_FIELD_1', 'RETREIVED_FIELD_2'],
        [[10,20], [30,40], [50, 60]],
    ),
    (
        "ContentAddressableMemory Func Initializer (regular 2d) Mech Default Variable Init",
        ContentAddressableMemory,
        {'initializer':np.array([[np.array([1,2]), np.array([3,4]), np.array([5, 6])],
                                 [[10,20], [30,40], [50,60]],
                                 [np.array([11,12]), np.array([22, 23]), np.array([34, 35])]])},
        {'default_variable':[[0,0],[0,0],[0,0]]},
        [[10,20], [30,40], [50, 60]],
        ['FIELD_0_INPUT', 'FIELD_1_INPUT', 'FIELD_2_INPUT'],
        ['RETREIVED_FIELD_0', 'RETREIVED_FIELD_1', 'RETREIVED_FIELD_2'],
        [[10,20], [30,40], [50, 60]],
    )
]

# Allows names to be with each test_data set
names = [test_data[i][0] for i in range(len(test_data))]

@pytest.mark.parametrize('name, func, func_params, mech_params, test_var,'
                         'input_port_names, output_port_names, expected_output', test_data, ids=names)
def test_with_contentaddressablememory(name, func, func_params, mech_params, test_var,
                                       input_port_names, output_port_names, expected_output, mech_mode):
    f = func(seed=0, **func_params)
    EpisodicMemoryMechanism(function=f, **mech_params)
    em = EpisodicMemoryMechanism(**mech_params)
    assert em.input_ports.names == input_port_names
    assert em.output_ports.names == output_port_names

    if mech_mode == 'Python':
        def EX(variable):
            em.execute(variable)
            return em.output_values
    elif mech_mode == 'LLVM':
        def EX(variable):
            # DUMMY UNTIL LLVM IMPLEMENTED
            return variable
    elif mech_mode == 'PTX':
        def EX(variable):
            # DUMMY UNTIL PTX IMPLEMENTED
            return variable

    EX(test_var)
    actual_output = EX(test_var)
    for i,j in zip(actual_output,expected_output):
        np.testing.assert_allclose(i, j, atol=1e-08)

def test_failures_with_contentaddressable_memory():

    # Initializer with >2d regular array
    with pytest.raises(FunctionError) as error_text:
        f = ContentAddressableMemory(initializer=[[[[1],[0],[1]], [[1],[0],[0]], [[0],[1],[1]]],
                                                  [[[0],[1],[0]], [[0],[1],[1]], [[1],[1],[0]]]])
        em = EpisodicMemoryMechanism(size = [1,1,1], function=f)
        em.execute([[[0],[1],[0]], [[0],[1],[1]], [[1],[1],[0]]])
    assert 'Attempt to store and/or retrieve an entry in ContentAddressableMemory ' \
           '([[[1]\n  [0]\n  [1]]\n\n [[1]\n  [0]\n  [0]]\n\n [[0]\n  [1]\n  [1]]]) ' \
           'that has more than 2 dimensions (3);  try flattening innermost ones.' in str(error_text.value)



from psyneulink import *
import numpy as np

# CONSTRUCT THE MODEL ***********************************

# Construct the color naming pathway:
color_input = ProcessingMechanism(name='COLOR INPUT', size=2) # Note:  default function is Linear
color_input_to_hidden_wts = np.array([[2, -2], [-2, 2]])
color_hidden = ProcessingMechanism(name='COLOR HIDDEN', size=2, function=Logistic(bias=-4))
color_hidden_to_output_wts = np.array([[2, -2], [-2, 2]])
output = ProcessingMechanism(name='OUTPUT', size=2, function=Logistic)
color_pathway = [color_input, color_input_to_hidden_wts, color_hidden, color_hidden_to_output_wts, output]

# Construct the word reading pathway (using the same output_layer)
word_input = ProcessingMechanism(name='WORD INPUT', size=2)
word_input_to_hidden_wts = np.array([[3, -3], [-3, 3]])
word_hidden = ProcessingMechanism(name='WORD HIDDEN', size=2, function=Logistic(bias=-4))
word_hidden_to_output_wts = np.array([[3, -3], [-3, 3]])
word_pathway = [word_input, word_input_to_hidden_wts, word_hidden, word_hidden_to_output_wts, output]

# Construct the task specification pathways
task_input = ProcessingMechanism(name='TASK INPUT', size=2)
task = LCAMechanism(name='TASK', size=2, initial_value=[0.5,0.5])
task_color_wts = np.array([[4,4],[0,0]])
task_word_wts = np.array([[0,0],[4,4]])
task_color_pathway = [task_input, task, task_color_wts, color_hidden]
task_word_pathway = [task_input, task, task_word_wts, word_hidden]

# Construct the decision pathway:
decision = DDM(name='DECISION', input_format=ARRAY)
decision_pathway = [output, decision]

# Construct control mechanism
control = ControlMechanism(name='CONTROL',
                           objective_mechanism=ObjectiveMechanism(name='Conflict Monitor',
                                                                  function=Energy(size=2,
                                                                                  matrix=[[0,-2.5],[-2.5,0]]),
                                                                  monitor=output),
                           default_allocation=[0.5],
                           control_signals=[(GAIN, task)])

# Construct the Composition:
Stroop_model = Composition(name='Stroop Model - Conflict Monitoring')
Stroop_model.add_linear_processing_pathway(color_pathway)
Stroop_model.add_linear_processing_pathway(word_pathway)
Stroop_model.add_linear_processing_pathway(task_color_pathway)
Stroop_model.add_linear_processing_pathway(task_word_pathway)
Stroop_model.add_linear_processing_pathway(decision_pathway)
Stroop_model.add_controller(control)

# Assign conditions:
settling_time = 10
Stroop_model.scheduler.add_condition(color_hidden, EveryNCalls(task, settling_time))
Stroop_model.scheduler.add_condition(word_hidden, EveryNCalls(task, settling_time))
Stroop_model.scheduler.add_condition(output,All(EveryNCalls(color_hidden, 1),
                                                           EveryNCalls(word_hidden, 1)))
Stroop_model.scheduler.add_condition(decision, EveryNCalls(output, 1))

# SHOW_GRAPH ***********************************
Stroop_model.show_graph(show_controller=True,
                        # show_node_structure=ALL
                        )

# RUN THE MODEL *********************************

# Label inputs
red = [1,0]
green = [0,1]
word = [0,1]
color = [1,0]

np.set_printoptions(precision=2)
global t
t = 0
def print_after():
    global t
    print(f'\nEnd of trial {t}:')
    print(f'\t\t\t\tcolor  word')
    print(f'\ttask:\t\t{task.value[0]}')
    print(f'\ttask gain:\t   {task.parameter_ports[GAIN].value}')
    # print(f'\ttask gain:\t   {task.function.parameters.gain}')
    print(f'\t\t\t\tred   green')
    print(f'\toutput:\t\t{output.value[0]}')
    print(f'\tdecision:\t{decision.value[0]}{decision.value[1]}')
    print(f'\tconflict:\t  {control.objective_mechanism.value[0]}')
    t += 1

task.log.set_log_conditions(VALUE)
control.log.set_log_conditions(VALUE)

task.initial_value = [0.5,0.5]
num_trials = 4
stimuli = {color_input:[red] * num_trials,
           word_input:[green] * num_trials,
           task_input:[color] * num_trials}

Stroop_model.run(inputs=stimuli,
                 # animate={'show_controller':True,
                 #          # 'show_cim':True
                 #          },
                 call_after_trial=print_after)

Stroop_model.log.print_entries(display=[TIME, VALUE])



import psyneulink as pnl

a = pnl.TransferMechanism()
b = pnl.TransferMechanism()
c = pnl.TransferMechanism()

p = pnl.Pathway(pathway=[a, b, c])

comp = pnl.Composition()
comp.add_backpropagation_learning_pathway(pathway=p)



from psyneulink import *

# TODO:
#   Nback::
#     - separate out stim/context external inputs from those from EM into FFN
#     - figure out how to specify feedback from DDM to EM
#     - figure out how to execute EM twice:
#       > first, at beginning of trial, to retrieve item based on current stimulus & context
#             (with prob retrieval = 1, prob storage = 0)
#       > second time, at end of trial (under influence of ControlMechanism) to encode current stimulus & context
#             (with prob storage = 1;  prob of retrieval = 0)
#     - implement circular drift as function for an input mechanism
#     - BUG:  should be able to use InputPort as spec for a pathway (if there is nothing after it);
#             same for OutputPort (if there is nothing before it)


#region N-BACK MODEL
def n_back_model():

    # Input Mechs
    stim = TransferMechanism(name='STIM', size=5)
    context = TransferMechanism(name='CONTEXT', size=5)

    # Feedforward Network:
    stim_input_layer = TransferMechanism(name='STIM INPUT LAYER', size=5)
    context_input_layer = TransferMechanism(name='CONTEXT INPUT LAYER', size=5)
    match_output_layer = TransferMechanism(name='MATCH LAYER', size=1)
    # ffn = AutodiffComposition(name='FFN', pathways=[[stim_input,match_output], [context_input, match_output]])
    ffn = Composition(name='FFN', pathways=[[stim_input_layer, match_output_layer],
                                            [context_input_layer, match_output_layer]])

    # Episodic Memory, Decision and Control
    # em = EpisodicMemoryMechanism(name='EM', content_size=5, assoc_size=5)
    em = EpisodicMemoryMechanism(name='EM', size=5,
                                 # function=DictionaryMemory(initializer=[[[0,0,0,0,0],[0,0,0,0,0]]])
                                 )
    ctl = ControlMechanism(control=(STORAGE_PROB, em))
    decision = DDM(name='DECISION')

    resp_decision = Pathway([match_output_layer, (decision, NodeRole.OUTPUT)])
    # FIX: ENHANCE add_linear_processing_pathway TO SUPPORT InputPort at end, or OutputPort at beginning:
    # stimulus_encoding = [stim, em.input_ports[KEY_INPUT]]
    # context_encoding = [context, em.input_ports[VALUE_INPUT]]

    # MappingProjection(sender=stim, receiver=stim_input_layer)
    # MappingProjection(sender=stim, receiver=em.input_ports[KEY_INPUT])
    # MappingProjection(sender=context, receiver=context_input_layer)
    # MappingProjection(sender=context, receiver=em.input_ports[VALUE_INPUT])
    # MappingProjection(sender=em.output_ports[KEY_OUTPUT], receiver=stim_input_layer)
    # MappingProjection(sender=em.output_ports[VALUE_OUTPUT], receiver=context_input_layer)
    # stim_processing = Pathway([stim, ffn])
    # context_processing = Pathway([context, ffn])
    # stim_encoding = Pathway([stim, em])
    # context_encoding = Pathway([context, em])
    # stim_retrieval = Pathway([em, stim_input_layer])
    # context_retrieval = Pathway([em, context_input_layer])
    # storage = Pathway([(decision, NodeRole.OUTPUT), (ctl, NodeRole.FEEDBACK_SENDER), em])
    # # FIX: show_graph NOT RECOGNIZING STIM->STIM_INPUT_LAYER AND CONTEXT->CONTEXT_INPUT_LAYER
    # comp = Composition(pathways=[stim_processing,
    #                              context_processing,
    #                              ffn,
    #                              context_encoding,
    #                              stim_encoding,
    #                              resp_decision,
    #                              stim_retrieval,
    #                              context_retrieval,
    #                              storage])
    # FIX: show_graph NOT RECOGNIZING STIM->STIM_INPUT_LAYER AND CONTEXT->CONTEXT_INPUT_LAYER
    # comp = Composition(pathways=[[stim, ffn],
    #                              [stim,em],
    #                              [context,ffn],
    #                              [context,em],
    #                              [em,ffn],
    #                              [ffn, em],
    #                              [ffn, decision, ctl, em]])

    # comp = Composition(pathways=[ffn,
    #                              [stim, stim_input_layer],
    #                              [stim, MappingProjection(stim, em.input_ports[KEY_INPUT]), em],
    #                              [context, context_input_layer],
    #                              [context, MappingProjection(context, em.input_ports[VALUE_INPUT]), em],
    #                              [em,stim_input_layer],
    #                              [em,context_input_layer],
    #                              [ffn, decision, ctl, em]])

    comp = Composition()
    comp.add_nodes([stim, context, ffn, em, (decision, NodeRole.OUTPUT), ctl])
    comp.add_projection(MappingProjection(), stim, stim_input_layer)
    comp.add_projection(MappingProjection(), context, context_input_layer)
    comp.add_projection(MappingProjection(), stim, em.input_ports[KEY_INPUT])
    comp.add_projection(MappingProjection(), context, em.input_ports[VALUE_INPUT])
    comp.add_projection(MappingProjection(), em.output_ports[KEY_OUTPUT], stim_input_layer)
    comp.add_projection(MappingProjection(), em.output_ports[VALUE_OUTPUT], context_input_layer)
    comp.add_projection(MappingProjection(), match_output_layer, decision)
    comp.add_projection(MappingProjection(), decision, ctl)
    # comp.add_projection(MappingProjection(), decision, stim_input_layer)

    # comp._analyze_graph()
    comp.show_graph()
    # comp.show_graph(show_cim=True,
    #                 show_node_structure=ALL,
    #                 show_projection_labels=True,
    #                 show_dimensions=True)
    # comp.show_graph(show_cim=True,
    #                 show_node_structure=ALL,
    #                 show_projection_labels=True,
    #                 show_dimensions=True)
    # comp.run(inputs={stim:[1,2,3,4,5],
    #                  context:[6,7,8,9,10]},
    #          report_output=ReportOutput.ON)
    # comp.run(inputs={a:2.5}, report_output=ReportOutput.FULL)
#endregion
n_back_model()



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* PNL LLVM helpers **************************************************************

from contextlib import contextmanager
from ctypes import util

from llvmlite import ir

from .debug import debug_env
from ..scheduling.condition import All, AllHaveRun, Always, Any, AtPass, AtTrial, BeforeNCalls, AtNCalls, AfterNCalls, \
    EveryNCalls, Never, Not, WhenFinished, WhenFinishedAny, WhenFinishedAll
from ..scheduling.time import TimeScale


@contextmanager
def for_loop(builder, start, stop, inc, id):
    # Initialize index variable
    assert start.type is stop.type
    index_var = builder.alloca(stop.type, name=id + "_index_var_loc")
    builder.store(start, index_var)

    # basic blocks
    cond_block = builder.append_basic_block(id + "-cond-bb")
    out_block = None

    # Loop condition
    builder.branch(cond_block)
    with builder.goto_block(cond_block):
        tmp = builder.load(index_var, name=id + "_cond_index_var")
        cond = builder.icmp_signed("<", tmp, stop, name=id + "_loop_cond")

        # Loop body
        with builder.if_then(cond, likely=True):
            index = builder.load(index_var, name=id + "_loop_index_var")

            yield (builder, index)

            index = builder.add(index, inc, name=id + "_index_var_inc")
            builder.store(index, index_var)
            builder.branch(cond_block)

        out_block = builder.block

    builder.position_at_end(out_block)


def for_loop_zero_inc(builder, stop, id):
    start = stop.type(0)
    inc = stop.type(1)
    return for_loop(builder, start, stop, inc, id)


def array_ptr_loop(builder, array, id):
    # Assume we'll never have more than 4GB arrays
    stop = ir.IntType(32)(array.type.pointee.count)
    return for_loop_zero_inc(builder, stop, id)


def fclamp(builder, val, min_val, max_val):
    min_val = min_val if isinstance(min_val, ir.Value) else val.type(min_val)
    max_val = max_val if isinstance(max_val, ir.Value) else val.type(max_val)

    cond = builder.fcmp_unordered("<", val, min_val)
    tmp = builder.select(cond, min_val, val)
    cond = builder.fcmp_unordered(">", tmp, max_val)
    return builder.select(cond, max_val, tmp)


def uint_min(builder, val, other):
    other = other if isinstance(other, ir.Value) else val.type(other)

    cond = builder.icmp_unsigned("<=", val, other)
    return builder.select(cond, val, other)


def get_param_ptr(builder, component, params_ptr, param_name):
    idx = ir.IntType(32)(component.llvm_param_ids.index(param_name))
    return builder.gep(params_ptr, [ir.IntType(32)(0), idx],
                       name="ptr_param_{}_{}".format(param_name, component.name))


def get_state_ptr(builder, component, state_ptr, stateful_name, hist_idx=0):
    idx = ir.IntType(32)(component.llvm_state_ids.index(stateful_name))
    ptr = builder.gep(state_ptr, [ir.IntType(32)(0), idx],
                      name="ptr_state_{}_{}".format(stateful_name,
                                                    component.name))
    # The first dimension of arrays is history
    if hist_idx is not None and isinstance(ptr.type.pointee, ir.ArrayType):
        assert len(ptr.type.pointee) > hist_idx, \
            "History not available: {} ({})".format(ptr.type.pointee, hist_idx)
        ptr = builder.gep(state_ptr, [ir.IntType(32)(0), idx,
                                      ir.IntType(32)(hist_idx)],
                          name="ptr_state_{}_{}_hist{}".format(stateful_name,
                                                               component.name,
                                                               hist_idx))
    return ptr


def push_state_val(builder, component, state_ptr, name, new_val):
    val_ptr = get_state_ptr(builder, component, state_ptr, name, None)
    for i in range(len(val_ptr.type.pointee) - 1, 0, -1):
        dest_ptr = get_state_ptr(builder, component, state_ptr, name, i)
        src_ptr = get_state_ptr(builder, component, state_ptr, name, i - 1)
        builder.store(builder.load(src_ptr), dest_ptr)

    dest_ptr = get_state_ptr(builder, component, state_ptr, name)
    builder.store(builder.load(new_val), dest_ptr)


def unwrap_2d_array(builder, element):
    if isinstance(element.type.pointee, ir.ArrayType) and isinstance(element.type.pointee.element, ir.ArrayType):
        assert element.type.pointee.count == 1
        return builder.gep(element, [ir.IntType(32)(0), ir.IntType(32)(0)])
    return element


def load_extract_scalar_array_one(builder, ptr):
    val = builder.load(ptr)
    if isinstance(val.type, ir.ArrayType) and val.type.count == 1:
        val = builder.extract_value(val, [0])
    return val


def fneg(builder, val, name=""):
    return builder.fsub(val.type(-0.0), val, name)


def exp(ctx, builder, x):
    exp_f = ctx.get_builtin("exp", [x.type])
    return builder.call(exp_f, [x])


def tanh(ctx, builder, x):
    tanh_f = ctx.get_builtin("tanh", [x.type])
    return builder.call(tanh_f, [x])


def coth(ctx, builder, x):
    coth_f = ctx.get_builtin("coth", [x.type])
    return builder.call(coth_f, [x])


def csch(ctx, builder, x):
    csch_f = ctx.get_builtin("csch", [x.type])
    return builder.call(csch_f, [x])


def is_close(ctx, builder, val1, val2, rtol=1e-05, atol=1e-08):
    is_close_f = ctx.get_builtin("is_close")
    rtol_val = val1.type(rtol)
    atol_val = val1.type(atol)
    return builder.call(is_close_f, [val1, val2, rtol_val, atol_val])


def all_close(ctx, builder, arr1, arr2, rtol=1e-05, atol=1e-08):
    assert arr1.type == arr2.type
    all_ptr = builder.alloca(ir.IntType(1))
    builder.store(all_ptr.type.pointee(1), all_ptr)
    with array_ptr_loop(builder, arr1, "all_close") as (b1, idx):
        val1_ptr = b1.gep(arr1, [idx.type(0), idx])
        val2_ptr = b1.gep(arr1, [idx.type(0), idx])
        val1 = b1.load(val1_ptr)
        val2 = b1.load(val2_ptr)
        res_close = is_close(ctx, b1, val1, val2, rtol, atol)

        all_val = b1.load(all_ptr)
        all_val = b1.and_(all_val, res_close)
        b1.store(all_val, all_ptr)

    return builder.load(all_ptr)


def create_allocation(builder, allocation, search_space, idx):
    # Construct allocation corresponding to this index
    for i in reversed(range(len(search_space.type.pointee))):
        slot_ptr = builder.gep(allocation, [idx.type(0), idx.type(i)])

        dim_ptr = builder.gep(search_space, [idx.type(0), idx.type(i)])
        # Iterators store {start, step, num}
        if isinstance(dim_ptr.type.pointee,  ir.LiteralStructType):
            iter_val = builder.load(dim_ptr)
            dim_start = builder.extract_value(iter_val, 0)
            dim_step = builder.extract_value(iter_val, 1)
            dim_size = builder.extract_value(iter_val, 2)
            dim_idx = builder.urem(idx, dim_size)
            val = builder.uitofp(dim_idx, dim_step.type)
            val = builder.fmul(val, dim_step)
            val = builder.fadd(val, dim_start)
        elif isinstance(dim_ptr.type.pointee,  ir.ArrayType):
            # Otherwise it's just an array
            dim_size = idx.type(len(dim_ptr.type.pointee))
            dim_idx = builder.urem(idx, dim_size)
            val_ptr = builder.gep(dim_ptr, [idx.type(0), dim_idx])
            val = builder.load(val_ptr)
        else:
            assert False, "Unknown dimension type: {}".format(dim_ptr.type)

        idx = builder.udiv(idx, dim_size)

        builder.store(val, slot_ptr)


def is_pointer(x):
    type_t = getattr(x, "type", x)
    return isinstance(type_t, ir.PointerType)

def is_floating_point(x):
    type_t = getattr(x, "type", x)
    # dereference pointer
    if is_pointer(type_t):
        type_t = type_t.pointee
    return isinstance(type_t, (ir.DoubleType, ir.FloatType, ir.HalfType))

def is_integer(x):
    type_t = getattr(x, "type", x)
    # dereference pointer
    if is_pointer(type_t):
        type_t = type_t.pointee
    return isinstance(type_t, ir.IntType)

def is_scalar(x):
    return is_integer(x) or is_floating_point(x)

def is_vector(x):
    type_t = getattr(x, "type", x)
    if is_pointer(type_t):
        type_t = type_t.pointee
    return isinstance(type_t, ir.ArrayType) and is_scalar(type_t.element)

def is_2d_matrix(x):
    type_t = getattr(x, "type", x)
    if is_pointer(type_t):
        type_t = type_t.pointee
    return isinstance(type_t, ir.ArrayType) and is_vector(type_t.element)

def is_boolean(x):
    type_t = getattr(x, "type", x)
    if is_pointer(type_t):
        type_t = type_t.pointee
    return isinstance(type_t, ir.IntType) and type_t.width == 1

def get_array_shape(x):
    x_ty = x.type
    if is_pointer(x):
        x_ty = x_ty.pointee

    assert isinstance(x_ty, ir.ArrayType), f"Tried to get shape of non-array type: {x_ty}"
    dimensions = []
    while hasattr(x_ty, "count"):
        dimensions.append(x_ty.count)
        x_ty = x_ty.element

    return dimensions

def array_from_shape(shape, element_ty):
    array_ty = element_ty
    for dim in reversed(shape):
        array_ty = ir.ArrayType(array_ty, dim)
    return array_ty

def recursive_iterate_arrays(ctx, builder, u, *args):
    """Recursively iterates over all elements in scalar arrays of the same shape"""
    assert isinstance(u.type.pointee, ir.ArrayType), "Can only iterate over arrays!"
    assert all(len(u.type.pointee) == len(v.type.pointee) for v in args), "Tried to iterate over differing lengths!"
    with array_ptr_loop(builder, u, "recursive_iteration") as (b, idx):
        u_ptr = b.gep(u, [ctx.int32_ty(0), idx])
        arg_ptrs = (b.gep(v, [ctx.int32_ty(0), idx]) for v in args)
        if is_scalar(u_ptr):
            yield (u_ptr, *arg_ptrs)
        else:
            yield from recursive_iterate_arrays(ctx, b, u_ptr, *arg_ptrs)

# TODO: Remove this function. Can be replaced by `recursive_iterate_arrays`
def call_elementwise_operation(ctx, builder, x, operation, output_ptr):
    """Recurse through an array structure and call operation on each scalar element of the structure. Store result in output_ptr"""
    for (inp_ptr, out_ptr) in recursive_iterate_arrays(ctx, builder, x, output_ptr):
        builder.store(operation(ctx, builder, builder.load(inp_ptr)), out_ptr)

def printf(builder, fmt, *args, override_debug=False):
    if "print_values" not in debug_env and not override_debug:
        return
    #FIXME: Fix builtin printf and use that instead of this
    try:
        import llvmlite.binding as llvm
        libc = util.find_library("c")
        llvm.load_library_permanently(libc)
        # Address will be none if the symbol is not found
        printf_address = llvm.address_of_symbol("printf")
    except Exception as e:
        return

    # Direct pointer constants don't work
    printf_ty = ir.FunctionType(ir.IntType(32), [ir.IntType(8).as_pointer()], var_arg=True)
    printf = builder.inttoptr(ir.IntType(64)(printf_address), printf_ty.as_pointer())
    ir_module = builder.function.module
    fmt += "\0"

    int8 = ir.IntType(8)
    fmt_data = bytearray(fmt.encode("utf8"))
    fmt_ty = ir.ArrayType(int8, len(fmt_data))
    global_fmt = ir.GlobalVariable(ir_module, fmt_ty,
                                   name="printf_fmt_" + str(len(ir_module.globals)))
    global_fmt.linkage = "internal"
    global_fmt.global_constant = True
    global_fmt.initializer = fmt_ty(fmt_data)

    fmt_ptr = builder.gep(global_fmt, [ir.IntType(32)(0), ir.IntType(32)(0)])
    builder.call(printf, [fmt_ptr] + list(args))


def printf_float_array(builder, array, prefix="", suffix="\n", override_debug=False):
    printf(builder, prefix, override_debug=override_debug)

    with array_ptr_loop(builder, array, "print_array_loop") as (b1, i):
        printf(b1, "%lf ", b1.load(b1.gep(array, [ir.IntType(32)(0), i])), override_debug=override_debug)

    printf(builder, suffix, override_debug=override_debug)


def printf_float_matrix(builder, matrix, prefix="", suffix="\n", override_debug=False):
    printf(builder, prefix, override_debug=override_debug)
    with array_ptr_loop(builder, matrix, "print_row_loop") as (b1, i):
        row = b1.gep(matrix, [ir.IntType(32)(0), i])
        printf_float_array(b1, row, suffix="\n", override_debug=override_debug)
    printf(builder, suffix, override_debug=override_debug)


class ConditionGenerator:
    def __init__(self, ctx, composition):
        self.ctx = ctx
        self.composition = composition
        self._zero = ctx.int32_ty(0) if ctx is not None else None

    def get_private_condition_struct_type(self, composition):
        time_stamp_struct = ir.LiteralStructType([self.ctx.int32_ty,   # Trial
                                                  self.ctx.int32_ty,   # Pass
                                                  self.ctx.int32_ty])  # Step

        status_struct = ir.LiteralStructType([
                    self.ctx.int32_ty,  # number of executions in this run
                    time_stamp_struct   # time stamp of last execution
                ])
        structure = ir.LiteralStructType([
            time_stamp_struct,  # current time stamp
            ir.ArrayType(status_struct, len(composition.nodes))  # for each node
        ])
        return structure

    def get_private_condition_initializer(self, composition):
        return ((0, 0, 0),
                tuple((0, (-1, -1, -1)) for _ in composition.nodes))

    def get_condition_struct_type(self, composition=None):
        composition = self.composition if composition is None else composition
        structs = [self.get_private_condition_struct_type(composition)]
        for node in composition.nodes:
            structs.append(self.get_condition_struct_type(node) if isinstance(node, type(self.composition)) else ir.LiteralStructType([]))
        return ir.LiteralStructType(structs)

    def get_condition_initializer(self, composition=None):
        composition = self.composition if composition is None else composition
        data = [self.get_private_condition_initializer(composition)]
        for node in composition.nodes:
            data.append(self.get_condition_initializer(node) if isinstance(node, type(self.composition)) else tuple())
        return tuple(data)

    def bump_ts(self, builder, cond_ptr, count=(0, 0, 1)):
        """
        Increments the time structure of the composition.
        Count should be a tuple where there is a number in only one spot, and zeroes elsewhere.
        Indices greater than that of the one are zeroed.
        """

        # Only one element should be non-zero
        assert count.count(0) == len(count) - 1

        # Get timestruct pointer
        ts_ptr = builder.gep(cond_ptr, [self._zero, self._zero, self._zero])
        ts = builder.load(ts_ptr)

        assert len(ts.type) == len(count)
        # Update run, pass, step of ts
        for idx in range(len(ts.type)):
            if all(v == 0 for v in count[:idx]):
                el = builder.extract_value(ts, idx)
                el = builder.add(el, self.ctx.int32_ty(count[idx]))
            else:
                el = self.ctx.int32_ty(0)
            ts = builder.insert_value(ts, el, idx)

        builder.store(ts, ts_ptr)
        return builder

    def ts_compare(self, builder, ts1, ts2, comp):
        assert comp == '<'

        # True if all elements to the left of the current one are equal
        prefix_eq = self.ctx.bool_ty(1)
        result = self.ctx.bool_ty(0)

        assert ts1.type == ts2.type
        for element in range(len(ts1.type)):
            a = builder.extract_value(ts1, element)
            b = builder.extract_value(ts2, element)

            # Use existing prefix_eq to construct expression
            # for the current element
            element_comp = builder.icmp_signed(comp, a, b)
            current_comp = builder.and_(prefix_eq, element_comp)
            result = builder.or_(result, current_comp)

            # Update prefix_eq
            element_eq = builder.icmp_signed('==', a, b)
            prefix_eq = builder.and_(prefix_eq, element_eq)

        return result

    def __get_node_status_ptr(self, builder, cond_ptr, node):
        node_idx = self.ctx.int32_ty(self.composition.nodes.index(node))
        return builder.gep(cond_ptr, [self._zero, self._zero, self.ctx.int32_ty(1), node_idx])

    def __get_node_ts(self, builder, cond_ptr, node):
        status_ptr = self.__get_node_status_ptr(builder, cond_ptr, node)
        ts_ptr = builder.gep(status_ptr, [self.ctx.int32_ty(0),
                                          self.ctx.int32_ty(1)])
        return builder.load(ts_ptr)

    def get_global_ts(self, builder, cond_ptr):
        ts_ptr = builder.gep(cond_ptr, [self._zero, self._zero, self._zero])
        return builder.load(ts_ptr)

    def generate_update_after_run(self, builder, cond_ptr, node):
        status_ptr = self.__get_node_status_ptr(builder, cond_ptr, node)
        status = builder.load(status_ptr)

        # Update number of runs
        runs = builder.extract_value(status, 0)
        runs = builder.add(runs, self.ctx.int32_ty(1))
        status = builder.insert_value(status, runs, 0)

        # Update time stamp
        ts = self.get_global_ts(builder, cond_ptr)
        status = builder.insert_value(status, ts, 1)

        builder.store(status, status_ptr)

    def generate_ran_this_pass(self, builder, cond_ptr, node):
        global_ts = self.get_global_ts(builder, cond_ptr)
        global_trial = builder.extract_value(global_ts, 0)
        global_pass = builder.extract_value(global_ts, 1)

        node_ts = self.__get_node_ts(builder, cond_ptr, node)
        node_trial = builder.extract_value(node_ts, 0)
        node_pass = builder.extract_value(node_ts, 1)

        pass_eq = builder.icmp_signed("==", node_pass, global_pass)
        trial_eq = builder.icmp_signed("==", node_trial, global_trial)
        return builder.and_(pass_eq, trial_eq)

    def generate_ran_this_trial(self, builder, cond_ptr, node):
        global_ts = self.get_global_ts(builder, cond_ptr)
        global_trial = builder.extract_value(global_ts, 0)

        node_ts = self.__get_node_ts(builder, cond_ptr, node)
        node_trial = builder.extract_value(node_ts, 0)

        return builder.icmp_signed("==", node_trial, global_trial)

    def generate_sched_condition(self, builder, condition, cond_ptr, node,
                                 is_finished_callbacks, num_exec_locs):


        if isinstance(condition, Always):
            return self.ctx.bool_ty(1)

        if isinstance(condition, Never):
            return self.ctx.bool_ty(0)

        elif isinstance(condition, Not):
            orig_condition = self.generate_sched_condition(builder, condition.condition, cond_ptr, node, is_finished_callbacks, num_exec_locs)
            return builder.not_(orig_condition)

        elif isinstance(condition, All):
            agg_cond = self.ctx.bool_ty(1)
            for cond in condition.args:
                cond_res = self.generate_sched_condition(builder, cond, cond_ptr, node, is_finished_callbacks, num_exec_locs)
                agg_cond = builder.and_(agg_cond, cond_res)
            return agg_cond

        elif isinstance(condition, AllHaveRun):
            # Extract dependencies
            dependencies = self.composition.nodes
            if len(condition.args) > 0:
                dependencies = condition.args

            run_cond = self.ctx.bool_ty(1)
            for node in dependencies:
                if condition.time_scale == TimeScale.TRIAL:
                    node_ran = self.generate_ran_this_trial(builder, cond_ptr, node)
                elif condition.time_scale == TimeScale.PASS:
                    node_ran = self.generate_ran_this_pass(builder, cond_ptr, node)
                else:
                    assert False, "Unsupported 'AllHaveRun' time scale: {}".format(condition.time_scale)
                run_cond = builder.and_(run_cond, node_ran)
            return run_cond

        elif isinstance(condition, Any):
            agg_cond = self.ctx.bool_ty(0)
            for cond in condition.args:
                cond_res = self.generate_sched_condition(builder, cond, cond_ptr, node, is_finished_callbacks, num_exec_locs)
                agg_cond = builder.or_(agg_cond, cond_res)
            return agg_cond

        elif isinstance(condition, AtTrial):
            trial_num = condition.args[0]
            global_ts = self.get_global_ts(builder, cond_ptr)
            trial = builder.extract_value(global_ts, 0)
            return builder.icmp_unsigned("==", trial, trial.type(trial_num))

        elif isinstance(condition, AtPass):
            pass_num = condition.args[0]
            global_ts = self.get_global_ts(builder, cond_ptr)
            current_pass = builder.extract_value(global_ts, 1)
            return builder.icmp_unsigned("==", current_pass,
                                         current_pass.type(pass_num))

        elif isinstance(condition, EveryNCalls):
            target, count = condition.args
            assert count == 1, "EveryNCalls isonly supprted with count == 1"

            target_ts = self.__get_node_ts(builder, cond_ptr, target)
            node_ts = self.__get_node_ts(builder, cond_ptr, node)

            # If target ran after node did its TS will be greater node's
            return self.ts_compare(builder, node_ts, target_ts, '<')

        elif isinstance(condition, BeforeNCalls):
            target, count = condition.args
            scale = condition.time_scale.value
            target_num_execs_in_scale = builder.gep(num_exec_locs[target],
                                                    [self.ctx.int32_ty(0),
                                                     self.ctx.int32_ty(scale)])
            num_execs = builder.load(target_num_execs_in_scale)

            return builder.icmp_unsigned('<', num_execs, self.ctx.int32_ty(count))

        elif isinstance(condition, AtNCalls):
            target, count = condition.args
            scale = condition.time_scale.value
            target_num_execs_in_scale = builder.gep(num_exec_locs[target],
                                                    [self.ctx.int32_ty(0),
                                                     self.ctx.int32_ty(scale)])
            num_execs = builder.load(target_num_execs_in_scale)
            return builder.icmp_unsigned('==', num_execs, self.ctx.int32_ty(count))

        elif isinstance(condition, AfterNCalls):
            target, count = condition.args
            scale = condition.time_scale.value
            target_num_execs_in_scale = builder.gep(num_exec_locs[target],
                                                    [self.ctx.int32_ty(0),
                                                     self.ctx.int32_ty(scale)])
            num_execs = builder.load(target_num_execs_in_scale)
            return builder.icmp_unsigned('>=', num_execs, self.ctx.int32_ty(count))

        elif isinstance(condition, WhenFinished):
            # The first argument is the target node
            assert len(condition.args) == 1
            target = is_finished_callbacks[condition.args[0]]
            is_finished_f = self.ctx.import_llvm_function(target[0], tags=frozenset({"is_finished", "node_wrapper"}))
            return builder.call(is_finished_f, target[1])

        elif isinstance(condition, WhenFinishedAny):
            assert len(condition.args) > 0

            run_cond = self.ctx.bool_ty(0)
            for node in condition.args:
                target = is_finished_callbacks[node]
                is_finished_f = self.ctx.import_llvm_function(target[0], tags=frozenset({"is_finished", "node_wrapper"}))
                node_is_finished = builder.call(is_finished_f, target[1])

                run_cond = builder.or_(run_cond, node_is_finished)

            return run_cond

        elif isinstance(condition, WhenFinishedAll):
            assert len(condition.args) > 0

            run_cond = self.ctx.bool_ty(1)
            for node in condition.args:
                target = is_finished_callbacks[node]
                is_finished_f = self.ctx.import_llvm_function(target[0], tags=frozenset({"is_finished", "node_wrapper"}))
                node_is_finished = builder.call(is_finished_f, target[1])

                run_cond = builder.and_(run_cond, node_is_finished)

            return run_cond

        assert False, "Unsupported scheduling condition: {}".format(condition)
